/*
 * @(#)FeatureDescriptor.java	1.15 97/01/03  
 * 
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion bdk_beta
 * Copyright (C) 1997 Microsoft Corporation. Portions copyright (C) 1993-1997 Sun Microsystems, Inc. All Rights Reserved.
 */

package java.beans;

import java.util.Hashtable;

/**
 * The FeatureDescriptor class is the common baseclass for PropertyDescriptor,
 * EventSetDescriptor, and MethodDescriptor, etc.
 * <p>
 * It supports some common information that can be set and retrieved for
 * any of the introspection descriptors.
 * <p>
 * In addition it provides an extension mechanism so that arbitrary
 * attribute/value pairs can be associated with a design feature.
 */

public class FeatureDescriptor {

    static final Object lock = new Object();

    public FeatureDescriptor() {
    }

    /**
     * @return The programmatic name of the property/method/event
     */
    public String getName() {
	return name;
    }

    /**
     * @param name  The programmatic name of the property/method/event
     */
    public void setName(String name) {
	this.name = name;
    }

    /**
     * @return The localized display name for the property/method/event.
     *	This defaults to the same as its programmatic name from getName.
     */
    public String getDisplayName() {
        String displayName = null;
        if (tearoffs != null) {
            displayName = (String) tearoffs[TEAROFF_DISPLAYNAME];
        }
        if (displayName == null) {
            return getName();
	}
        return displayName;
    }

    /**
     * @param displayName  The localized display name for the
     *		property/method/event.
     */
    public void setDisplayName(String displayName) {
        ensureTearOffsExist();
        tearoffs[TEAROFF_DISPLAYNAME] = displayName;
    }

    /**
     * The "expert" flag is used to distinguish between those features that are
     * intended for expert users from those that are intended for normal users.
     *
     * @return True if this feature is intended for use by experts only.
     */
    public boolean isExpert() {
        return (featureFlags & FEATURE_EXPERT) != 0;
    }

    /**
     * The "expert" flag is used to distinguish between features that are
     * intended for expert users from those that are intended for normal users.
     *
     * @param expert True if this feature is intended for use by experts only.
     */
    public void setExpert(boolean expert) {
        twiddleFeatureBit(expert, FEATURE_EXPERT);
    }

    /**
     * The "hidden" flag is used to identify features that are intended only
     * for tool use, and which should not be exposed to humans.
     *
     * @return True if this feature should be hidden from human users.
     */
    public boolean isHidden() {
        return (featureFlags & FEATURE_HIDDEN) != 0;
    }

    /**
     * The "hidden" flag is used to identify features that are intended only
     * for tool use, and which should not be exposed to humans.
     *
     * @param hidden  True if this feature should be hidden from human users.
     */
    public void setHidden(boolean hidden) {
        twiddleFeatureBit(hidden, FEATURE_HIDDEN);
    }

    /**
     * @return  A localized short description associated with this 
     *   property/method/event.  This defaults to be the display name.
     */
    public String getShortDescription() {
        String shortDescription = null;
        if (tearoffs != null) {
            shortDescription = (String) tearoffs[TEAROFF_SHORTDESCRIPTION];
        }
        if (shortDescription == null) {
	    return getDisplayName();
	}
	return shortDescription;
    }

    /**
     * You can associate a short descriptive string with a feature.  Normally
     * these descriptive strings should be less than about 40 characters.
     * @param text  A (localized) short description to be associated with
     * this property/method/event.
     */
    public void setShortDescription(String text) {
        ensureTearOffsExist();
        tearoffs[TEAROFF_SHORTDESCRIPTION] = text;
    }

    /**
     * Associate a named attribute with this feature.
     * @param attributeName  The locale-independent name of the attribute
     * @param value  The value.
     */
    public void setValue(String attributeName, Object value) {
        ensureTearOffTableExists();
        ((Hashtable) tearoffs[TEAROFF_TABLE]).put(attributeName, value);
    }

    /**
     * Retrieve a named attribute with this feature.
     * @param attributeName  The locale-independent name of the attribute
     * @return  The value of the attribute.  May be null if
     *	   the attribute is unknown.
     */
    public Object getValue(String attributeName) {
        if (tearoffs == null || tearoffs[TEAROFF_TABLE] == null) {
            return null;
	}
        return ((Hashtable) tearoffs[TEAROFF_TABLE]).get(attributeName);
    }

    /**
     * @return  An enumeration of the locale-independent names of any 
     *    attributes that have been registered with setValue.
     */
    public java.util.Enumeration attributeNames() {
        ensureTearOffTableExists();
        return ((Hashtable) tearoffs[TEAROFF_TABLE]).keys();
    }

    /**
     * Package-private constructor,
     * Merge information from two FeatureDescriptors.
     * The merged hidden and expert flags are formed by or-ing the values.
     * In the event of other conflicts, the second argument (y) is
     * given priority over the first argument (x).
     * @param x  The first (lower priority) MethodDescriptor
     * @param y  The second (higher priority) MethodDescriptor
     */
    FeatureDescriptor(FeatureDescriptor x, FeatureDescriptor y) {
        featureFlags = (x.featureFlags & (FEATURE_EXPERT | FEATURE_HIDDEN)) |
                       (y.featureFlags & (FEATURE_EXPERT | FEATURE_HIDDEN)) |
                       (featureFlags & ~(FEATURE_EXPERT | FEATURE_HIDDEN));
        name = y.name;

        if (x.tearoffs != null || y.tearoffs != null) {

            ensureTearOffsExist();

            if (y.tearoffs != null) {
                tearoffs[TEAROFF_SHORTDESCRIPTION] =
                    y.tearoffs[TEAROFF_SHORTDESCRIPTION];
            }
            if (tearoffs[TEAROFF_SHORTDESCRIPTION] == null) {
                if (x.tearoffs != null) {
                    tearoffs[TEAROFF_SHORTDESCRIPTION] =
                        x.tearoffs[TEAROFF_SHORTDESCRIPTION];
                }
            }

            if (y.tearoffs != null) {
                tearoffs[TEAROFF_DISPLAYNAME] =
                    y.tearoffs[TEAROFF_DISPLAYNAME];
            }
            if (tearoffs[TEAROFF_DISPLAYNAME] == null) {
                if (x.tearoffs != null) {
                    tearoffs[TEAROFF_DISPLAYNAME] =
                        x.tearoffs[TEAROFF_DISPLAYNAME];
                }
            }

            if (x.tearoffs != null) {
                addTable((Hashtable) x.tearoffs[TEAROFF_TABLE]);
            }
            if (y.tearoffs != null) {
                addTable((Hashtable) y.tearoffs[TEAROFF_TABLE]);
            }
        }
    }

    private void addTable(java.util.Hashtable t) {
	if (t == null) {
	    return;
	}
	java.util.Enumeration keys = t.keys();
	while (keys.hasMoreElements()) {
	    String key = (String)keys.nextElement();
	    Object value = t.get(key);
	    setValue(key, value);
	}
    }

    void ensureTearOffsExist() {
        synchronized(lock) {
            if (tearoffs == null) {
                tearoffs = new Object[TEAROFF_COUNT];
            }
        }
    }

    void ensureTearOffTableExists() {
        ensureTearOffsExist();
        synchronized(lock) {
            if (tearoffs[TEAROFF_TABLE] == null) {
                tearoffs[TEAROFF_TABLE] = new Hashtable(1);
            }
        }
    }

    void twiddleFeatureBit(boolean on, int bit) {
        //  Atomically update the featureFlags field.  This may be overkill, but
        //  it guarantees compatible behavior with the JDK where each of the
        //  feature bits were seperate boolean fields.  This 'twiddling' is rare
        //  so we avoid allocating a monitor for every FeatureDescriptor that
        //  needs to do this.
        synchronized(lock) {
            if (on) {
                featureFlags |= bit;
            } else {
                featureFlags &= (~bit);
            }
        }
    }

    final static int FEATURE_EXPERT         = 0x00000001;
    final static int FEATURE_HIDDEN         = 0x00000002;
    final static int FEATURE_BOUND          = 0x00000004;
    final static int FEATURE_CONSTRAINED    = 0x00000008;
    int featureFlags;

    private String name;

    final static int TEAROFF_SHORTDESCRIPTION   = 0;
    final static int TEAROFF_DISPLAYNAME        = 1;
    final static int TEAROFF_TABLE              = 2;
    final static int TEAROFF_EXTERNAL           = 3;
    final static int TEAROFF_COUNT              = 4;
    Object tearoffs[];
}
