/*
 * @(#)ObjectInputStream.java	1.30 97/05/02
 * 
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.1_beta
 * Copyright (C) 1997 Microsoft Corporation. Portions copyright (C) 1993-1997 Sun Microsystems, Inc. All Rights Reserved.
 */

package java.io;

import java.lang.reflect.Array;
import java.util.Vector;

/**
 * An ObjectInputStream deserializes primitive data and objects previously
 * written using an ObjectOutputStream.
 * 
 * ObjectOutputStream and ObjectInputStream can provide an application
 * with persistent storage for graphs of objects when used with a
 * FileOutputStream and FileInputStream respectively.
 * ObjectInputStream is used to recover those objects previously
 * serialized. Other uses include passing objects between hosts using
 * a socket stream or for marshaling and unmarshaling arguments and
 * parameters in a remote communication system.<p>
 *
 * ObjectInputStream ensures that the types of all objects in the
 * graph created from the stream match the classes present in the
 * Java Virtual Machine.  Classes are loaded as required using the
 * standard mechanisms. <p>
 *
 * Only objects that support the java.io.Serializable or
 * java.io.Externalizable interface can be read from streams.
 *
 * The method <STRONG>readObject</STRONG> is used to read an object
 * from the stream.  Java's safe casting should be used to get the
 * desired type.  In Java, strings and arrays are objects and are
 * treated as objects during serialization. When read they need to be
 * cast to the expected type.<p>
 *
 * Primitive data types can be read from the stream using the appropriate
 * method on DataInput. <p>
 * 
 * The default deserialization mechanism for objects restores the
 * contents of each field to the value and type it had when it was written.
 * Fields declared as transient or static are ignored by the
 * deserialization process.  References to other objects cause those
 * objects to be read from the stream as necessary.  Graphs of objects
 * are restored correctly using a reference sharing mechanism.  New
 * objects are always allocated when deserializing, which prevents
 * existing objects from being overwritten. <p>
 *
 * Reading an object is analogous to running the constructors of a new
 * object.  Memory is allocated for the object and initialized to zero
 * (NULL).  No-arg constructors are invoked for the non-serializable
 * classes and then the fields of the serializable classes are
 * restored from the stream starting with the serializable class closest to
 * java.lang.Object and finishing with the object's most specifiec
 * class. <p>
 *
 * For example, to read from a stream as written by the example in
 * ObjectOutputStream: <br>
 *
 * <PRE>
 *	FileInputStream istream = new FileInputStream("t.tmp");
 *	ObjectInputStream p = new ObjectInputStream(istream);
 *
 *	int i = p.readInt();
 *	String today = (String)p.readObject();
 *	Date date = (Date)p.readObject();
 *
 *	istream.close();
 * </PRE>
 *
 * Classes control how they are serialized by implementing either the
 * java.io.Serializable or java.io.Externalizable interfaces.<P>
 *
 * Implementing the Serializable interface allows object serialization
 * to save and restore the entire state of the object and it allows
 * classes to evolve between the time the stream is written and the time it is
 * read.  It automatically traverses references between objects,
 * saving and restoring entire graphs.
 *
 * Serializable classes that require special handling during the
 * serialization and deserialization process should implement both
 * of these methods:<p>
 *
 * <PRE>
 * private void writeObject(java.io.ObjectOutputStream stream)
 *     throws IOException;
 * private void readObject(java.io.ObjectInputStream stream)
 *     throws IOException, ClassNotFoundException; 
 * </PRE><p>
 *
 * The readObject method is responsible for reading and restoring the
 * state of the object for its particular class using data written to
 * the stream by the corresponding writeObject method.  The method
 * does not need to concern itself with the state belonging to its
 * superclasses or subclasses.  State is restored by reading data from
 * the ObjectInputStream for the individual fields and making
 * assignments to the appropriate fields of the object.  Reading
 * primitive data types is supported by DataInput. <p>
 *
 * Serialization does not read or assign values to the fields of any
 * object that does not implement the java.io.Serializable interface.
 * Subclasses of Objects that are not serializable can be
 * serializable. In this case the non-serializable class must have a
 * no-arg constructor to allow its fields to be initialized.  In this
 * case it is the responsibility of the subclass to save and restore
 * the state of the non-serializable class. It is frequently the case that
 * the fields of that class are accessible (public, package, or
 * protected) or that there are get and set methods that can be used
 * to restore the state. <p>
 *
 * Any exception that occurs while deserializing an object will be
 * caught by the ObjectInputStream and abort the reading process. <p>
 *
 * Implementing the Externalizable interface allows the object to
 * assume complete control over the contents and format of the object's
 * serialized form.  The methods of the Externalizable interface,
 * writeExternal and readExternal, are called to save and restore the
 * objects state.  When implemented by a class they can write and read
 * their own state using all of the methods of ObjectOutput and
 * ObjectInput.  It is the responsibility of the objects to handle any
 * versioning that occurs.
 *
 * @see java.io.DataInput
 * @see java.io.ObjectOutputStream
 * @see java.io.Serializable
 * @since   JDK1.1
 */

public class ObjectInputStream 
	extends InputStream 
	implements ObjectInput, ObjectStreamConstants 
{	
	private InputStream         m_is;
	private DataInputStream     m_dis;

	// map between the object and the handle assigned to it
	private Object[]            m_aObjAndHandleMap;
	private int                 m_nObjHandleMapSize;

    // info about current read
	private Object              m_objCurrent;
	private ObjectStreamClass   m_descCurrent;

    // flag to enable subclasses to resolve objects after they're read
	private boolean             m_bResolveEnabled;

	// keeps track of the recursive calls to readObject()
	private int                 m_nRecursionLevel;

	// flag to start or stop buffering
    private boolean             m_bCurrentBufferingMode;

	// number of bytes to follow in the stream
	private int                 m_nBlockSize;

	// bytes left in the block data record
	private int                 m_nBlockBytesLeft;

	// header that identifies the type of the next object in stream
	private byte                m_headerCurrent;

	// vector to hold objects and priorities for validating
	private Vector              m_vValidationCallbacks;
    
	// Store the descriptors and corresponding classes
	private ObjectStreamClass[] m_aClassDescriptors;
	private Class[]             m_aClasses; 
	private int                 m_nStackTop;

	private Object              m_objNullOrPrev;

    // native methods
    private native void invokeDefaultReadObject( Object o, Class cl )
		throws InvalidClassException, StreamCorruptedException,
								ClassNotFoundException, IOException;
    

    private native void invokeReadObject( Object o, Class cl )
		throws InvalidClassException, StreamCorruptedException,
								ClassNotFoundException, IOException;

	private native Object allocateNewObject(Class aclass, Class initclass)
	    throws InstantiationException, IllegalAccessException, NoSuchMethodError;

    /* Resolve a class name relative to the specified class.  If the
     * class is null find the first available class loader up the
     * stack.  This will resolve classes relative to the caller of
     * ObjectInputStream instead of the itself. Classes must be
     * loaded/resolved relative to the application.
     */
    private native Class loadClass0( Class cl, String classname )
	    throws ClassNotFoundException;
	

/**
 * Create an ObjectInputStream that reads from the specified InputStream.
 * The stream header containing the magic number and version number
 * are read from the stream and verified. This method will block
 * until the corresponding ObjectOutputStream has written and flushed the header.
 * @exception StreamCorruptedException The version or magic number are incorrect.
 * @exception IOException An exception occurred in the underlying stream.
 * @since     JDK1.1
 */
public ObjectInputStream( InputStream IS ) 
	throws IOException, StreamCorruptedException
{
	m_is = IS;
	m_dis = new DataInputStream(this);

    // Check the stream header 
	readStreamHeader();

	setDefaultStreamValues();
}


/**
 * Read an object from the ObjectInputStream.
 * The class of the object, the signature of the class, and the values
 * of the non-transient and non-static fields of the class and all
 * of its supertypes are read.  Default deserializing for a class can be
 * overriden using the writeObject and readObject methods.
 * Objects referenced by this object are read transitively so
 * that a complete, equivalent graph of objects is reconstructed by readObject. <p>
 *
 * The root object is completly restored when all of its fields
 * and the objects it references are completely restored.  At this
 * point the object validation callbacks are executed in order
 * based on their registered priorities. The callbacks are
 * registered by objects (in the readObject special methods)
 * as they are individually restored.
 *
 * Exceptions are thrown for problems with the InputStream and for classes
 * that should not be deserialized.  All exceptions are fatal to the 
 * InputStream and leave it in an indeterminate state; it is up to the caller
 * to ignore or recover the stream state.
 * @exception java.lang.ClassNotFoundException Class of a serialized object
 *      cannot be found.
 * @exception InvalidClassException Something is wrong with a class used by
 *     serialization.
 * @exception StreamCorruptedException Control information in the
 *     stream is inconsistent.
 * @exception OptionalDataException Primitive data was found in the 
 * stream instead of objects.
 * @exception IOException Any of the usual Input/Output related exceptions.
 * @since     JDK1.1
 */
public final Object readObject() 
	throws OptionalDataException, ClassNotFoundException, IOException 
{
	// Check whether the first thing is a block data record.
    // If yes, then we have to throw a OptionalDataException
	if( m_bCurrentBufferingMode && ((m_nBlockBytesLeft = available()) > 0) ) 
	    throw new OptionalDataException( m_nBlockBytesLeft );
    
	// Get the byte that represents a valid header byte (See ObjectStreamConstants)
	// Reset the stream if needed.
    byte    header = readHeader();

	// Remember the current state
	Object                  objPrevious = m_objCurrent;
	ObjectStreamClass       descPrevious = m_descCurrent;
    Object                  objFinal = null;
	IOException             exFatalIOException = null;
	ClassNotFoundException  exFatalClassNotFoundException = null;

    // Stop reading in bufferred mode
	boolean bOldBufferingMode = toggleBufferingMode( false );

    m_nRecursionLevel++;
	
	try 
	{  	 
		// Check if the stream has a null or a previous reference  
		if( checkNullOrPreviousReference() )
			return m_objNullOrPrev;

		// Check if the stream has a array, string, class or class descriptor 
		if( checkSpecialCases() )
			return( objFinal = m_objCurrent );

		int nBlockSize;

        // Take further action based on the type of the code 
        switch( header )
		{ 
		case TC_OBJECT: // read the new object 
			int     nHandle  = readNewObject();

			objFinal = m_objCurrent;

			// Resolve the object. This is done by the sub classes of the stream 
			if( m_bResolveEnabled )
			{
				// resolve the object and remember the resolve object in place
                //  of the original 
				objFinal = resolveObject( objFinal );
				m_aObjAndHandleMap[ nHandle ] = objFinal;
			}
		    break; 

		case TC_BLOCKDATA:
	    case TC_BLOCKDATALONG:
            if( header == TC_BLOCKDATALONG ) 
			{
                // Size of the block is in the next 32 bits
                int b3 = m_is.read();
                int b2 = m_is.read();
                int b1 = m_is.read();
                int b0 = m_is.read();

	            if( (b3 | b2 | b1 | b0) < 0 )
                    throw new StreamCorruptedException( "EOF expecting count" );

                nBlockSize = (b3 << 24) | (b2 << 16) | (b1 << 8) | b0;
                if( nBlockSize < 0 )
                    throw new StreamCorruptedException( "Negative block data size" );
			}
            else
			{
                // Size of the block is in the next byte
                nBlockSize = m_is.read();
                if( nBlockSize < 0 )
                    throw new StreamCorruptedException( "EOF expecting count" );
			}

            // Block data should occur only in blocking mode 
            if( !bOldBufferingMode )
                throw new StreamCorruptedException( "Unexpected blockdata" );
		
            throw new OptionalDataException( nBlockSize );

		case TC_ENDBLOCKDATA:
		    if( !bOldBufferingMode )
                throw new StreamCorruptedException( "Unexpected end of block data" );

            throw new OptionalDataException( true );

        case TC_EXCEPTION:
            setDefaultStreamValues();   // reset the stream context  
            objFinal = readObject();    // read in the exception object 

            // reset the stream context so that reading begins afresh the next time 
            //  readObject is called
            setDefaultStreamValues();

            throw new WriteAbortedException( "Write aborted exception", (Exception)objFinal ); 
		
        default :
		    throw new StreamCorruptedException( "Unknown code in readObject " + header );
		}
	} 
	catch( OptionalDataException exOptionalData )
	{
        // This is not a noop. It's thrown here so that the caller can catch it and 
		// skip over optional data. If it is not thrown here, it will be caought as an 
		// IOException and meaningful information will be lost.
		throw exOptionalData;
	}
	catch( IOException exIO )
	{
		if( (exFatalIOException == null) && (exFatalClassNotFoundException == null) )
		    exFatalIOException = exIO;
	}
	catch( ClassNotFoundException exClassNotFound )
	{
		if( (exFatalIOException == null) && (exFatalClassNotFoundException == null) )
			exFatalClassNotFoundException = exClassNotFound;
	}
	finally
	{
        // wind down in case this is a recursive call
        m_nRecursionLevel--;
        m_objCurrent = objPrevious; 
        m_descCurrent = descPrevious; 
        toggleBufferingMode( bOldBufferingMode );
	}

    // Rethrow exceptions if needed 
	IOException exIO = exFatalIOException;
	if( m_nRecursionLevel == 0 )
	    exFatalIOException = null;
	if( null != exIO )
	    throw exIO;

	ClassNotFoundException  exClassNotFound = exFatalClassNotFoundException;
	if( m_nRecursionLevel == 0 )
	    exFatalClassNotFoundException = null;
	if( exClassNotFound != null )
	    throw exClassNotFound;
	
    // validate the graph of objects if we're finished reading
	if( m_nRecursionLevel == 0 )
		validateGraph();

    return objFinal; 
} 

/**
 * Read the non-static and non-transient fields of the current class
 * from this stream.  This may only be called from the readObject method
 * of the class being deserialized. It will throw the NotActiveException
 * if it is called otherwise.
 *
 * @exception java.lang.ClassNotFoundException if the class of a serialized
 *              object could not be found.
 * @exception IOException        if an I/O error occurs.
 * @exception NotActiveException if the stream is not currently reading
 *              objects.
 * @since     JDK1.1
 */
public final void defaultReadObject() 
	throws IOException, ClassNotFoundException, NotActiveException
{
	// Make sure that the method is called only from readObject method 
	if( (m_objCurrent == null) || (m_descCurrent == null) )
		throw new NotActiveException("Invalid call to defaultReadObject");

    // save old buffering mode
	boolean bOldBufferingMode = toggleBufferingMode( false );

	if( m_descCurrent.isVersionMismatched() )  
	{
	    // versions do not match, do a special read
        m_descCurrent.doMismatchedRead( this, m_objCurrent );
	}
    else
    {
        // versions match, call regular default reader
	    invokeDefaultReadObject( m_objCurrent, m_descCurrent.forClass() );
    }

	toggleBufferingMode( bOldBufferingMode );
} 


/**
 * Register an object to be validated before the graph is
 * returned.  While similar to resolveObject these validations are
 * called after the entire graph has been reconstituted.
 * Typically, a readObject method will register the object with
 * the stream so that when all of the objects are restored a final
 * set of validations can be performed.
 * @param obj the object to receive the validation callback.
 * @param nPriority controls the order of callbacks;zero is a good default.
 * Use higher numbers to be called back earlier, lower numbers for later
 * callbacks. Within a priority, callbacks are processed in no
 * particular order.
 *
 * @exception NotActiveException The stream is not currently reading objects
 * so it is invalid to register a callback.
 * @exception InvalidObjectException The validation object is null.
 * @since     JDK1.1
 */
public synchronized void registerValidation( ObjectInputValidation obj, int nPriority ) 
	throws NotActiveException, InvalidObjectException
{

	if( m_nRecursionLevel == 0 )
	    throw new NotActiveException("readObject not Active");

	if( obj == null )
	    throw new InvalidObjectException("Null is not a valid callback object");

	ObjAndPriority  oap = new ObjAndPriority( obj, nPriority );

	if( m_vValidationCallbacks == null )
	{
	    m_vValidationCallbacks = new Vector( 128, 128 );
	}

	// Standard insert in a priority queue
	if( m_vValidationCallbacks.isEmpty() ||
	        (((ObjAndPriority)(m_vValidationCallbacks.lastElement())).m_nPriority >= nPriority) )
	{
		m_vValidationCallbacks.addElement( oap );
	    return;
	}

	int nSize = m_vValidationCallbacks.size();
	for( int index = 0; index < nSize; index++ ) 
	{	     
	    if( ((ObjAndPriority)m_vValidationCallbacks.elementAt( index )).m_nPriority <= nPriority )
		{
			m_vValidationCallbacks.insertElementAt( oap, index );
			break;
		}
	}		
} 

/////////////////////////////////////////////////////////////////////////////////////////////////

///////////////// DataInput Methods///////////////////////////////////////////////////////////////
    
/**
 * Reads a byte of data. This method will block if no input is 
 * available.
 * @return 	the byte read, or -1 if the end of the
 *		stream is reached.
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public int read()
	throws IOException 
{
	int nByte;
	if( m_bCurrentBufferingMode )
	{
	    if( m_nBlockSize == 0 )
			refill();
	    if( m_nBlockSize < 0 ) // EOF 
			return -1;

	    nByte = m_is.read();
	    if( nByte >= 0 )
			m_nBlockSize--;
	} 
	else 
	    nByte = m_is.read();		

	return nByte;
}

/**
 * Reads into an array of bytes.  This method will
 * block until some input is available.
 * @param data the buffer into which the data is read
 * @param nOffset the start offset of the data
 * @param nLength the maximum number of bytes read
 * @return  the actual number of bytes read, -1 is
 * 		returned when the end of the stream is reached.
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public int read( byte[] data, int nOffset, int nLength ) 
	throws IOException 
{

	if( nLength < 0 )
	    throw new IndexOutOfBoundsException();

	if( m_bCurrentBufferingMode )
	{
		int i;
	    if( 0 == m_nBlockSize )
			refill();
	    if( m_nBlockSize< 0 )
			return -1;

	    int j = Math.min( nLength, m_nBlockSize );

	    i = m_is.read( data, nOffset, j );
	    if( i > 0 )
			m_nBlockSize -= i;

	    return i;			// number of bytes read 
	}
	else 
	{
	    // read directly from input stream
	    return m_is.read( data, nOffset, nLength );
	}
}

/**
 * Returns the number of bytes that can be read without blocking.
 * @return the number of available bytes.
 * @since     JDK1.1
 */
public int available() 
	throws IOException 
{
    // If the blocking mode is on then give the number of bytes in the block data record.
    //  (This helps in skipping over block data, if needed.)
    //  Otherwise, return 0.
	if( m_bCurrentBufferingMode ) 
	{
	    if( (m_nBlockSize == 0) && (m_is.available() > 0) )
			refill();
	    if( m_nBlockSize >= 0 ) 
			return m_nBlockSize;
		else
			return 0;	// EOF is no bytes available
	} 
	else 
	    return 0;		// Not blockdata, no bytes available 
}

/**
 * Closes the input stream. Must be called
 * to release any resources associated with
 * the stream.
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */	
public void close() 
	throws IOException 
{
	m_is.close(); 
} 
/////////////////////////////////////////////////////////////////////////////////////////////////

/////////// DataInputStream methods are implemented here. They are simply redirected calls.///////

/**
 * Reads in a boolean.
 * @return the boolean read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public boolean readBoolean() 
	throws IOException 
{
	return m_dis.readBoolean(); 
} 

/**
 * Reads an 8 bit byte.
 * @return the 8 bit byte read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public byte readByte()
	throws IOException
{
	return m_dis.readByte(); 
} 

/**
 * Reads an unsigned 8 bit byte.
 * @return the 8 bit byte read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public int readUnsignedByte()
	throws IOException 
{
	return m_dis.readUnsignedByte(); 
} 

/**
 * Reads a 16 bit short.
 * @return the 16 bit short read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public short readShort()
	throws IOException
{
	return m_dis.readShort(); 
} 

/**
 * Reads an unsigned 16 bit short.
 * @return the 16 bit short read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public int readUnsignedShort()
	throws IOException
{
	return m_dis.readUnsignedShort(); 
} 

/**
 * Reads a 16 bit char.
 * @return the 16 bit char read. 
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public char readChar()
	throws IOException
{
	return m_dis.readChar(); 
} 

/**
 * Reads a 32 bit int.
 * @return the 32 bit integer read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public int readInt() 
	throws IOException 
{
	return m_dis.readInt(); 
} 

/**
 * Reads a 64 bit long.
 * @return the read 64 bit long.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public long readLong() 
	throws IOException 
{
	return m_dis.readLong(); 
} 

/**
 * Reads a 32 bit float.
 * @return the 32 bit float read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public float readFloat() 
	throws IOException 
{
	return m_dis.readFloat(); 
} 

/**
 * Reads a 64 bit double.
 * @return the 64 bit double read.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public double readDouble() 
	throws IOException 
{
	return m_dis.readDouble(); 
}

/**
 * Reads bytes, blocking until all bytes are read.
 * @param b	the buffer into which the data is read
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public void readFully( byte[] data )
	throws IOException
{
	m_dis.readFully( data ); 
} 

/**
 * Reads bytes, blocking until all bytes are read.
 * @param data the buffer into which the data is read
 * @param nOffset the start offset of the data
 * @param nSize the maximum number of bytes to read
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */	
public void readFully( byte[] data, int nOffset, int nSize ) 
	throws IOException
{
	if( nSize < 0 ) 
		throw new IndexOutOfBoundsException();

	m_dis.readFully( data, nOffset, nSize );
} 

/**
 * Skips bytes, block until all bytes are skipped.
 * @param nLength the number of bytes to be skipped
 * @return	the actual number of bytes skipped.
 * @exception EOFException If end of file is reached.
 * @exception IOException If other I/O error has occurred.
 * @since     JDK1.1
 */
public int skipBytes( int nLength )
	throws IOException
{
	return m_dis.skipBytes( nLength );
}

/**
 * Reads in a line that has been terminated by a \n, \r, 
 * \r\n or EOF.
 * @return a String copy of the line.
 * @since     JDK1.1
 */
public String readLine() 
	throws IOException 
{
	return m_dis.readLine(); 
} 

/**
 * Reads a UTF format String.
 * @return the String.
 * @since     JDK1.1
 */
public String readUTF() 
	throws IOException 
{
	return m_dis.readUTF(); 
}

/////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////Protected methods are defined here/////////////////////////////////////

/**
 * Subclasses may implement this method to allow classes to be
 * fetched from an alternate source. 
 *
 * The corresponding method in ObjectOutputStream is
 * annotateClass.  This method will be invoked only once for each
 * unique class in the stream.  This method can be implemented by
 * subclasses to use an alternate loading mechanism but must
 * return a Class object.  Once returned, the serialVersionUID of the
 * class is compared to the serialVersionUID of the serialized class.
 * If there is a mismatch, the deserialization fails and an exception
 * is raised. <p>
 *
 * By default the class name is resolved relative to the class
 * that called readObject. <p>
 *
 * @exception ClassNotFoundException If class of
 * a serialized object cannot be found.
 * @since     JDK1.1
 */
protected Class resolveClass( ObjectStreamClass descriptor )
	throws IOException, ClassNotFoundException 
{
	return loadClass0( null, descriptor.getName() );			 
} 

/**
 * This method will allow trusted subclasses of ObjectInputStream
 * to substitute one object for another during
 * deserialization. Replacing objects is disabled until
 * enableResolveObject is called. The enableResolveObject method
 * checks that the stream requesting to resolve object can be
 * trusted. Every reference to serializable objects is passed to
 * resolveObject.  To insure that the private state of objects is
 * not unintentionally exposed only trusted streams may use
 * resolveObject. <p>
 *
 * This method is called after an object has been read but before it is
 * returned from readObject.  The default resolveObject method
 * just returns the new object. <p>
 *
 * When a subclass is replacing objects it must insure that the
 * substituted object is compatible with every field where the
 * reference will be stored.  Objects whose type is not a subclass
 * of the type of the field or array element abort the
 * serialization by raising an exception and the object is not be
 * stored. <p>
 *
 * This method is called only once when each object is first encountered.
 * All subsequent references to the object will be redirected to the
 * new object. <P>
 *
 * @exception IOException Any of the usual Input/Output exceptions.
 * @since     JDK1.1
 */
protected Object resolveObject( Object obj )
	throws IOException 
{
	return obj;
}  

/**
 * Enable the stream to allow objects read from the stream to be replaced.
 * If the stream is a trusted class it is allowed to enable replacment.
 * Trusted classes are those classes with a classLoader equals null. <p>
 * 
 * When enabled the resolveObject method is called for every object
 * being deserialized.
 * 
 * @exception SecurityException The classloader of this stream object is non-null.
 * @since     JDK1.1
 */
protected final boolean enableResolveObject( boolean bEnable )
	throws SecurityException
{
	boolean bOldEnableState = m_bResolveEnabled;

	if( bEnable )
	{
		ClassLoader loader = this.getClass().getClassLoader();
		if( loader == null )
		{
			m_bResolveEnabled = true;
			return bOldEnableState;
		}

	    throw new SecurityException( "Not trusted class" );
	} 
	else 
	{
        m_bResolveEnabled = false;
	}
	return bOldEnableState;
}	

/**
 * The readStreamHeader method is provided to allow subclasses to
 * read and verify their own stream headers. It reads and
 * verifies the magic number and version number.
 * @since     JDK1.1
 */
protected void readStreamHeader()  
	throws IOException, StreamCorruptedException 
{		
	if( readShort() != STREAM_MAGIC )
	    throw new StreamCorruptedException( "InputStream does not contain a serialized object" );

    short   sVersion = readShort();

	if( sVersion != STREAM_VERSION )
	    throw new StreamCorruptedException( "Version Mismatch, Expected " +
					                            STREAM_VERSION + " and got " + sVersion );
}

/////////// These are private helper methods/////////////////////////////////////////////// 

// Call validate object on each object in the order of priority
private void validateGraph() 
	throws InvalidObjectException
{
	if( m_vValidationCallbacks == null )
		return;

	int nSize = m_vValidationCallbacks.size();

	for( int index = 0; index < nSize; index++ )
		((ObjAndPriority)m_vValidationCallbacks.elementAt( index )).m_obj.validateObject();
}

// Read the next header from the stream. If the header has not been read so far, get it from
//  the input stream
private byte readHeader() 
    throws IOException, StreamCorruptedException
{
	while( m_headerCurrent == 0 )
	{
        // read header byte from input stream 
        int   header = m_is.read();
        
        if( 0 > header ) 
		    throw new EOFException( "Invalid Header" + header );
	    
        m_headerCurrent = (byte)header;

        if( (m_headerCurrent < TC_BASE) || (m_headerCurrent > TC_MAX) )
		    throw new StreamCorruptedException( "Header out of range " + m_headerCurrent );

	    // Reset the stream if needed 
	    if( m_headerCurrent == TC_RESET) 
		{
            if( (m_objCurrent != null) || (m_descCurrent != null) || (m_nRecursionLevel != 0) ) 
		        throw new StreamCorruptedException( "Reset occurred while reading" );

		   // reset the stream 
	       setDefaultStreamValues();
	    }
	}

	return m_headerCurrent; 
}


private void resetHeader() 
{ 
	m_headerCurrent = 0x00; 
}


// If the object in the stream is null or a previous reference return with null or 
//  the previuously constructed object
private boolean checkNullOrPreviousReference() 
	throws IOException 
{  
	switch( readHeader() )
	{ 
    case TC_NULL: 
	    m_objNullOrPrev = null;
		resetHeader();
        return true;

	case TC_REFERENCE:
	    resetHeader();

        // read the refernence handle
		int   nHandle = readInt() - baseWireHandle;

        // throw an exception if handle is out of bounds
		if( (nHandle > m_nObjHandleMapSize) || (nHandle < 0) ) 
		    throw new StreamCorruptedException( "Reference to object never serialized" );

        // get the object from our handle->object map
        m_objNullOrPrev = m_aObjAndHandleMap[ nHandle ];
        return true;

    default:
        return false;
	}
}

// Check if the object is a string, array, class or a objectstream descriptor. 
// Instantiate a object of proper type and then return
private boolean checkSpecialCases()
	throws IOException, ClassNotFoundException, StreamCorruptedException 
{
	switch( readHeader() )
	{ 
	case TC_STRING: 
		m_objCurrent = readString();
		return true;

	case TC_ARRAY:
		m_objCurrent = readArray();
		return true;

	case TC_CLASSDESC:
		m_objCurrent = readClassDescriptor();
		return true;

	case TC_CLASS:
		m_objCurrent = readClass();
		return true;

	default:
		return false;
	}        
}

private Object readString()  
	throws IOException
{
	resetHeader();

    // read in the string in UTF format
    Object  objFinal = (Object)readUTF();

	// assign a new handle 
	makeNewHandle( objFinal );

	return objFinal; 
}

private Object readArray()  
	throws IOException, StreamCorruptedException, ClassNotFoundException 
{ 
    resetHeader(); 
	readHeader(); 

	// read in the class descriptor for the array 
	ObjectStreamClass   descriptor = readClassDescriptor();

	if( descriptor == null )
		throw new ClassNotFoundException( "Class descriptor not found in stream" );

	// read in the length of the array 
	int nLength = readInt();

	if( nLength < 0 )
		throw new StreamCorruptedException( "Invalid array length in stream" );

	// read in the values of the array
	Object  array = readArrayValues( descriptor, nLength );

	return array;
}

private Object readClass() 
	throws IOException, ClassNotFoundException 
{  
    resetHeader(); 
    readHeader(); 

	// Now read in the class descriptor and make sure it's non-null
	ObjectStreamClass   descriptor = readClassDescriptor();

	if( descriptor == null )
		throw new ClassNotFoundException( "Class descriptor not found in stream" );

    // get the class descriptor's class
	Object  objFinal = descriptor.forClass();

	// assign a new handle
	makeNewHandle( objFinal );

	return objFinal;
}

private ObjectStreamClass readClassDescriptor()
	throws IOException, ClassNotFoundException 
{
    // Check for null or previous reference 
	if( checkNullOrPreviousReference() )
	{
		return( (ObjectStreamClass)m_objNullOrPrev );
	}

    if( TC_CLASSDESC != readHeader() ) 
        throw new StreamCorruptedException( "Invalid Header" );

	resetHeader(); 

	// Read the name and SUID
	String  strClassName = readUTF(); 
	long    lSerialVersion = readLong();			
	
	// Construct the object stream class descriptor 
    ObjectStreamClass descriptor = new ObjectStreamClass( strClassName, lSerialVersion );
	
	makeNewHandle( descriptor );
    descriptor.read( this );		
  
	boolean bOldBufferingMode = toggleBufferingMode( true );

	// resolve the class
	Class   classLocal = resolveClass( descriptor );

	goToEndOfBlockData();
	toggleBufferingMode( bOldBufferingMode );
    
	if( null == classLocal )
		throw new ClassNotFoundException( descriptor.getName() );

	descriptor.setClass( classLocal );
		
    // Prepare to read in the superclass descriptor 
	readHeader(); 	
	descriptor.setSuperClass( readClassDescriptor() );
	
	return descriptor;
}
  
private int readNewObject()
	throws IOException, ClassNotFoundException, InvalidClassException
{		
	int nHandle = -1;

	resetHeader();      // consume TC_OBJECT 
	readHeader(); 

	// read in the class descriptor 
	if( (m_descCurrent = readClassDescriptor()) == null ) 
	    throw new ClassNotFoundException( "Class descriptor not found in stream" );

	// If Externalizable, create an instance and tell it to read its data.
	//  else, handle it as a serializable class.
	if( m_descCurrent.isExternalizable() )
	{
        // create a new object and assign it a handle
		m_objCurrent = initNewObject( m_descCurrent.forClass(), m_descCurrent.forClass() );
		nHandle = makeNewHandle( m_objCurrent );

        // ask the object to read itself
		((Externalizable)m_objCurrent).readExternal( this );
    } 
	else
	{		
		/* Now we have all the descriptors for the Object in stream. 
		 * Compare the classes of these descriptors with the local class 
		 * hierarchy. The descriptor,class pair is stored for each descriptor 
		 * where class is the corresponding class in the local hierarchy. If there
		 * is no such class then we store null. 
		 */
		ObjectStreamClass   descCurrent = m_descCurrent;
		Class               classCurrent = m_descCurrent.forClass();
		int                 nPrevStackTop = m_nStackTop;

		// Note that the classes of the object are processed from supertype to subtype 
		// Match the descriptor list against the local class hierarchy
		for( ; descCurrent != null; descCurrent = descCurrent.getSuperclass() )
		{
			Class   classLocal = classCurrent;
			Class   classDesc = descCurrent.forClass();
		
			for( ; classLocal != null; classLocal = classLocal.getSuperclass() )
			{
				if( classLocal == classDesc ) 
				{
                // found a superclass that matches this descriptor
				    break;
				}

				// Otherwise this local class was added. 
				// It will be initialized to the default values.
			}

			// Test if there is room for this new entry.
 			// If not, double the size of the arrays and copy the contents.
			m_nStackTop++;
			if( m_nStackTop >= m_aClasses.length )
			{
				int                 nNewLen = m_aClasses.length * 2;
				Class[]             aNewClasses = new Class[ nNewLen ];
				ObjectStreamClass[] aNewClassDescs = new ObjectStreamClass[ nNewLen ];

				System.arraycopy( m_aClasses, 0, aNewClasses, 0, m_aClasses.length );
				System.arraycopy( m_aClassDescriptors, 0, aNewClassDescs, 0, m_aClasses.length );

				m_aClasses = aNewClasses;
				m_aClassDescriptors = aNewClassDescs;
			}

			if( classLocal == null )
			{
				// No class was found corresponding to this descriptor
				m_aClassDescriptors[ m_nStackTop ] = descCurrent;
				m_aClasses[ m_nStackTop ] = null;
			} 
			else 
			{
       			// Yahoo ! There is a class matching the current descriptor class. 
				m_aClassDescriptors[ m_nStackTop ] = descCurrent;
				m_aClasses[ m_nStackTop ] = classLocal;
				classCurrent = classLocal.getSuperclass();
			}
		}
	    
		try 
		{
			// Create a new instance of the class. The highest serializable class is
			//  initialized and the sub classes are set to default values.  assign
            //  a handle to this new object.
			m_objCurrent = initNewObject( m_descCurrent.forClass(), classCurrent );
			nHandle = makeNewHandle( m_objCurrent );
	  
			// Now read in the data for all the classes present in the stream 
			for( ; m_nStackTop > nPrevStackTop ; m_nStackTop-- )
			{
				m_descCurrent = m_aClassDescriptors[ m_nStackTop ];

				if( m_aClasses[ m_nStackTop ] != null )
				{
                    toggleBufferingMode( true );

					// There is a class corresponding to the descriptor
					if( m_descCurrent.hasReadObject() )
					{
                        // ask the class to read itself
						invokeReadObject( m_objCurrent, m_descCurrent.forClass() );
					}
					else
                    {
                        // perform the default read operation
						defaultReadObject();
                    }
				}
				else 
				{
                    // There is no class corresponding to this Descriptor,
                    //  skip over the fields
					boolean bOldBufferingMode = toggleBufferingMode( false );

					m_descCurrent.skipFieldData( this );
					toggleBufferingMode( bOldBufferingMode );
				}

                 // If the source class (stream) had a write object method
                 //  it may have written more data and will have written the
                 //  TC_ENDBLOCKDATA.  Skip anything up to that and read it.
				if( m_descCurrent.hasWriteObject() )
					goToEndOfBlockData();

                toggleBufferingMode( false );
			}
		}
		finally 
		{
            // Reset to the previous top of stack 
		    m_nStackTop = nPrevStackTop;
		}

	}  
	return nHandle;
}

private Object initNewObject( Class classCurrent, Class classHighest )
	throws InvalidClassException
{
    Object  obj;

    try
    {
        obj = allocateNewObject( classCurrent, classHighest );
    }
    catch( NoSuchMethodError e )
    {
		throw new InvalidClassException( classCurrent.getName(),
                                                        "NoSuchMethodError" );
    }
    catch( IllegalAccessException e )
    {
		throw new InvalidClassException( classCurrent.getName(),
                                                        "IllegalAccessException" );
    }
    catch( InstantiationException e )
    {
		throw new InvalidClassException( classCurrent.getName(),
					                                    "InstantiationException");
    }

    return obj;
}

// Generate a New Handle 
private int makeNewHandle(Object obj) 
	throws IOException
{
	// Extend the array if there isn't room for this new element
	if( m_nObjHandleMapSize == m_aObjAndHandleMap.length ) 
	{
		Object[]    aOldHandles = m_aObjAndHandleMap;

		m_aObjAndHandleMap = new Object[ m_nObjHandleMapSize * 2 ];
	    System.arraycopy( aOldHandles, 0, m_aObjAndHandleMap, 0, m_nObjHandleMapSize );
	}

	m_aObjAndHandleMap[ m_nObjHandleMapSize ] = obj;

	return m_nObjHandleMapSize++;
}

// Restore the initial state of the stream 
private void setDefaultStreamValues() 
	throws IOException 
{
	m_objCurrent = null;
	m_descCurrent = null; 
	m_aObjAndHandleMap = new Object[ 128 ]; 
	m_nObjHandleMapSize   = 0; 
	m_bResolveEnabled  = false; 
	m_nRecursionLevel = 0;
	m_headerCurrent = 0x00; 
	m_aClasses = new Class[ 16 ];
	m_aClassDescriptors = new ObjectStreamClass[ 16 ];
	m_nStackTop = -1;
	m_objNullOrPrev = null;
	toggleBufferingMode( true );
}

// This method changes the buffering mode.  	
private boolean toggleBufferingMode( boolean bNewMode) 
	throws IOException 
{
	if( m_bCurrentBufferingMode == bNewMode ) 
		return bNewMode;
	
	if( m_bCurrentBufferingMode &&  (m_nBlockSize > 0) )
		throw new StreamCorruptedException( "Unread data" );

    // Set count to allow reading or not
	m_nBlockSize = (bNewMode ? 0 : -1);
	m_bCurrentBufferingMode = bNewMode; 

    // The previous mode is returned  
	return !bNewMode;
}

// Skip over any unread block data till TC_ENDBLOCKDATA
private void goToEndOfBlockData()
	throws IOException, ClassNotFoundException 
{
	while( TC_ENDBLOCKDATA != readHeader() )
	{
        try 
        {
            readObject(); 
        } 
        catch( OptionalDataException e ) 
        {
            if( e.length > 0 )
	            skip( e.length );  
        }
	}

	resetHeader();
}



// A data block should occur in the stream. Read the length of the block. 
// The data will be read by the caller. If error or EOF occurs return -1 
private void refill() 
	throws IOException
{
	byte    headerNext;

	m_nBlockSize= -1;	//  No more data to read, EOF

	try 
	{
	    headerNext  = readHeader();
	} 
	catch( EOFException e )
	{
	    return;
	}

	if( TC_BLOCKDATA == headerNext )
	{
	    resetHeader();
	    int nNewBlockCount = m_is.read();

	    if( (nNewBlockCount & 0xFFFFFF00) != 0 )
		    throw new StreamCorruptedException( "EOF expecting count" );

        m_nBlockSize = nNewBlockCount;
	} 
	else if( TC_BLOCKDATALONG == headerNext ) 
	{
	    resetHeader();

	    int b3 = m_is.read();
	    int b2 = m_is.read();
	    int b1 = m_is.read();
	    int b0 = m_is.read();
	    if( (b3 | b2 | b1 | b0) < 0 )
			throw new StreamCorruptedException( "Invalid length of block data" );

	    m_nBlockSize = (b3 << 24) | (b2 << 16) | (b1 << 8) | b0;

	    if( m_nBlockSize < 0 )
			throw new StreamCorruptedException( "Negative block data size" );
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////


/////////Methods to handle arrays////////////////////////////////////////////////////////////////

private Object readArrayValues( ObjectStreamClass descriptor, int nLength )
		throws IOException, ClassNotFoundException
{ 
	// Get the type of the array from the Descriptor and then call the 
	//  type specific method to read the array values.
	switch( descriptor.getName().charAt( 1 ) ) 
	{ 
		case 'I': return readIntArray( nLength );
		case 'B': return readByteArray( nLength );
		case 'S': return readShortArray( nLength );	       
		case 'J': return readLongArray( nLength ); 
		case 'F': return readFloatArray( nLength ); 
		case 'D': return readDoubleArray( nLength );
		case 'C': return readCharArray( nLength ); 
		case 'Z': return readBooleanArray( nLength );

		case '[': 
		case 'L': return readObjectArray( descriptor, nLength );

		default: throw new StreamCorruptedException( "Unknown Array Type" );
	}
}

private Object readIntArray( int nLength )
	throws IOException
{
	int[] aInts = new int[ nLength ];

	makeNewHandle( aInts );
	for( int index = 0; index < aInts.length ; index++ )
		aInts[ index ] = readInt();

	return aInts;
}

private Object readByteArray( int nLength )
	throws IOException 
{ 
	byte[]  aBytes = new byte[ nLength ];

	makeNewHandle( aBytes );
	for( int index = 0; index < aBytes.length; index++ )  
		aBytes[ index ] = readByte();

	return aBytes; 
}


private Object readShortArray( int nLength )
	throws IOException 
{ 
	short[] aShorts = new short[ nLength ];

	makeNewHandle( aShorts );
	for( int index = 0; index < aShorts.length; index++ )  
		aShorts[ index ] = readShort();

	return aShorts;
}

private Object readLongArray( int nLength )
		throws IOException 
{ 
	long[]  aLongs = new long[ nLength ];

	makeNewHandle( aLongs );
	for( int index = 0; index < aLongs.length; index++ )
		aLongs[ index ] = readLong();

	return aLongs; 
}

private Object readFloatArray( int nLength )
	throws IOException 
{ 
	float[] aFloats = new float[ nLength ];

	makeNewHandle( aFloats );
	for( int index = 0; index < aFloats.length; index++ )
		aFloats[ index ] = readFloat();

	return aFloats;
}

private Object readDoubleArray( int nLength )
	throws IOException 
{ 
	double[]    aDoubles = new double[ nLength ];

	makeNewHandle( aDoubles );
	for( int index = 0; index < aDoubles.length; index++ )  
		aDoubles[ index ] = readDouble();

	return aDoubles;
}

private Object readCharArray( int nLength )
	throws IOException 
{ 
	char[]  aChars = new char[ nLength ];

	makeNewHandle( aChars );
	for( int index = 0; index < aChars.length; index++ )
		aChars[ index ] = readChar();

	return aChars; 
}

private Object readBooleanArray( int nLength )
	throws IOException 
{ 
	boolean[] aBools = new boolean[ nLength ]; 

	makeNewHandle( aBools );
	for( int index = 0; index < aBools.length; index++ )  
		aBools[ index ] = readBoolean();

	return aBools; 
}


private Object readObjectArray( ObjectStreamClass descriptor, int nLength ) 
	throws IOException, ClassNotFoundException 
{
    Object[]    objArray = (Object[])Array.newInstance(
                                        descriptor.forClass().getComponentType(),
                                        nLength );

	makeNewHandle( objArray );
	for( int index = 0; index < nLength; index++)
		objArray[ index ] = readObject();

    return objArray;
}

}   // end of class ObjectInputStream

//////////////////////////////////////////////////////////////////////////////////////////////////

// Class for holding objects and priority
class ObjAndPriority 
{
	ObjectInputValidation   m_obj;
	int                     m_nPriority;

	ObjAndPriority( ObjectInputValidation obj, int nPriority )
	{
		m_obj = obj;
		m_nPriority = nPriority;
	}
}

