/*
 * @(#)ObjectOutputStream.java	1.29 97/05/02
 *
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * CopyrightVersion 1.1_beta
 * Copyright (C) 1997 Microsoft Corporation. Portions copyright (C) 1993-1997 Sun Microsystems, Inc. All Rights Reserved.
 */

package java.io;

import java.util.Hashtable;
import com.ms.util.OrdinalMap;

/**
 * An ObjectOutputStream writes primitive data types and graphs of
 * Java objects to an OutputStream.  The objects can be read
 * (reconstituted) using an ObjectInputStream.
 * Persistent storage of objects can be accomplished by using a file for
 * the stream.
 * If the stream is a network socket stream, the objects can be reconsituted
 * on another host or in another process. <p>
 *
 * Only objects that support the java.io.Serializable interface can be
 * written to streams.
 *
 * The class of each serializable object is encoded including the class
 * name and signature of the class, the values of the
 * object's fields and arrays, and the closure of any other objects
 * referenced from the initial objects. <p>
 *
 * The method <STRONG>writeObject</STRONG> is used to write an object
 * to the stream.  Any object, including Strings and arrays, is
 * written with writeObject. Multiple objects or primitives can be
 * written to the stream.  The objects must be read back from the
 * corresponding ObjectInputstream with the same types and in the same
 * order as they were written.<p>
 *
 * Primitive data types can also be written to the stream using the
 * appropriate methods from DataOutput. Strings can also be written
 * using the writeUTF method.<p>
 *
 * The default serialization mechanism for an object writes the class
 * of the object, the class signature, and the values of all
 * non-transient and non-static fields.  References to other objects
 * (except in transient or static fields) cause those objects to be
 * written also. Multiple references to a single object are encoded
 * using a reference sharing mechanism so that graphs of objects can
 * be restored to the same shape as when the original was written. <p>
 *
 * For example to write an object that can be read by the example in ObjectInputStream: <br>
 * <PRE>
 *	FileOutputStream ostream = new FileOutputStream("t.tmp");
 *	ObjectOutputStream p = new ObjectOutputStream(ostream);
 *
 *	p.writeInt(12345);
 *	p.writeObject("Today");
 *	p.writeObject(new Date());
 *
 *	p.flush();
 *	ostream.close();
 *
 * </PRE>
 *
 * Classes that require special handling during the serialization and deserialization
 * process must implement special methods with these exact signatures: <p>
 *
 * <PRE>
 * private void readObject(java.io.ObjectInputStream stream)
 *     throws IOException, ClassNotFoundException;
 * private void writeObject(java.io.ObjectOutputStream stream)
 *     throws IOException
 * </PRE><p>
 * The writeObject method is responsible for writing the state of
 * the object for its particular class so that the corresponding
 * readObject method can restore it.
 * The method does not need to concern itself with the
 * state belonging to the object's superclasses or subclasses.
 * State is saved by writing the individual fields to the ObjectOutputStream
 * using the writeObject method or by using the methods for
 * primitive data types supported by DataOutput. <p>
 *
 * Serialization does not write out the fields of any object that does
 * not implement the java.io.Serializable interface.  Subclasses of
 * Objects that are not serializable can be serializable. In this case
 * the non-serializable class must have a no-arg constructor to allow
 * its fields to be initialized.  In this case it is the
 * responsibility of the subclass to save and restore the state of the
 * non-serializable class. It is frequently the case that the fields
 * of that class are accessible (public, package, or protected) or
 * that there are get and set methods that can be used to restore the
 * state. <p>
 *
 * Serialization of an object can be prevented by implementing writeObject
 * and readObject methods that throw the NotSerializableException.
 * The exception will be caught by the ObjectOutputStream and abort the
 * serialization process.
 *
 * Implementing the Externalizable interface allows the object to
 * assume complete control over the contents and format of the object's
 * serialized form.  The methods of the Externalizable interface,
 * writeExternal and readExternal, are called to save and restore the
 * objects state.  When implemented by a class they can write and read
 * their own state using all of the methods of ObjectOutput and
 * ObjectInput.  It is the responsibility of the objects to handle any
 * versioning that occurs.
 *
 * @see java.io.DataOutput
 * @see java.io.ObjectInputStream
 * @see java.io.Serializable
 * @see java.io.Externalizable
 * @since       JDK1.1
 */

//////Implementation of the ObjectOutputStream Class. ///////////////////////////

public class ObjectOutputStream
	extends    OutputStream
	implements ObjectOutput, ObjectStreamConstants
{
    private OutputStream        m_out;
	private DataOutputStream    m_dos;

    // Keeps track of the nesting level on the stack
	private int                 m_nRecursionLevel;

    // Current object that is being written
    private Object              m_objCurrent;

    // Descriptor of current object or one of it's superclasses
	private ObjectStreamClass   m_descCurrent;

    // maps objects to their assigned handle
	private	OrdinalMap          m_mapObjectToHandle;

    // Flag for enabling object replacement
	private boolean             m_bReplaceEnabled;

  	// Map between original and replaced objects
	private Hashtable           m_htOriginalToAlternateObject;

 	// Leaves the stream in a unknown state
    private IOException         m_exFatalException;

	// block buffering data
	private boolean             m_bCurrentBufferingMode;
    private byte[]              m_aBlockData;
    private int                 m_nBlockSize;


    // native methods
    private native void invokeDefaultWriteObject( Object o, Class cl )
	    throws IOException, InvalidClassException;

    private native void invokeWriteObject( Object o, Class cl )
		throws IOException, InvalidClassException;
	

/**
 * Creates an ObjectOutputStream that writes to the specified OutputStream.
 * The stream header is written to the stream. The caller may want to call
 * flush immediately so that the corresponding ObjectInputStream can read
 * the header immediately.
 * @exception IOException Any exception thrown by the underlying OutputStream.
 * @since     JDK1.1
 */
public ObjectOutputStream( OutputStream out )
	throws IOException
{
	m_out = out;
	m_dos = new DataOutputStream(this);
	m_aBlockData = new byte[255];

	writeStreamHeader();
	setDefaultStreamValues();
			
}

/**
 * Write the specified object to the ObjectOutputStream.
 * The class of the object, the signature of the class, and the values
 * of the non-transient and non-static fields of the class and all
 * of its supertypes are written.  Default serialization for a class can be
 * overridden using the writeObject and the readObject methods.
 * Objects referenced by this object are written transitively so
 * that a complete equivalent graph of objects can be
 * reconstructed by an ObjectInputStream.  <p>
 *
 * Exceptions are thrown for
 * problems with the OutputStream and for classes that should not be
 * serialized.  All exceptions are fatal to the OutputStream, which
 * is left in an indeterminate state, and it is up to the caller
 * to ignore or recover the stream state.
 * @exception InvalidClassException Something is wrong with a class used by
 *	   serialization.
 * @exception NotSerializableException Some object to be serialized does not
 *	  implement the java.io.Serializable interface.
 * @exception IOException Any exception thrown by the underlying OutputStream.
 * @since     JDK1.1
 */
public final void writeObject( Object obj )
	throws IOException
{
    // This will drain any bufferred data
	boolean             bOldBufferingMode = toggleBufferingMode( false );
	Object              objPrevious = m_objCurrent;
	ObjectStreamClass   descPrevious = m_descCurrent;

	m_nRecursionLevel++;

	try
	{
		// A Null object or a object already serialized is handled by the following call
		if( checkNullAndPreviousReference( obj ) )
			return;

		// Check if the object is a class, string or array or ObjectStreamDescriptor
		if( checkSpecialCases( obj ) )
			return;

		// Check if replacement is enabled. If yes, then allow the
		//  subclasses to replace the object.
		if( m_bReplaceEnabled )
		{
			Object  objAlternate = replaceObject( obj );

			if( obj != objAlternate )
			{
                // Gotta replace the current object
				if( (objAlternate != null) &&
                            !(objAlternate instanceof java.io.Serializable) )
				{
					throw new NotSerializableException( objAlternate.getClass().getName() );
				}

				// This object needs to be replaced. Create the
				//  mapping between this object and it's replacement
				mapOriginalToAlternate( obj, objAlternate );

				// Check if the alternate object is already serialized or null
				//  If yes, then return immediately
				if( checkNullAndPreviousReferenceNoReplace( objAlternate ) )
					return;

				// Check if the object is a array, class, string or ObjectStreamDescriptor
				if( checkSpecialCases( objAlternate ) )
					return;

				// Now the alternate object becomes the current object to be written
				obj = objAlternate;
			}
		}
		
        writeNewObject( obj );
	}
	catch( ObjectStreamException exObjectStreamException )
	{
		// A NotSerializable or NotActive exception was thrown
		if( m_exFatalException == null )
		{
            // set the exception now to keep from recurring indefinitely if writing
            //  the exception to the stream throws an exception
            m_exFatalException = exObjectStreamException;

            try
			{
				// try writing the exception to the stream
                toggleBufferingMode( false );
				writeHeader( TC_EXCEPTION );
				setDefaultStreamValues();
				writeObject( exObjectStreamException );

				// reset the stream again so that no back reference to this exception occurs
				setDefaultStreamValues();
				m_exFatalException = exObjectStreamException;
			}
		    catch( IOException exDeadly )
			{
                if( exDeadly != (IOException)exObjectStreamException )
                {
            		// Horrible !! Exception occurred while writing the exception
			    	m_exFatalException = new StreamCorruptedException( exDeadly.getMessage() );
                }
		    }
        }

    }
	catch( IOException exDeadly )
	{
	    // Fatal IOException occurred - If one has already occurred then that is the
		//  one which is thrown again
        if( m_exFatalException == null )
            m_exFatalException = exDeadly;
	}
    finally
	{
		// wind down in case this is a recursive call
		m_nRecursionLevel--;
		m_objCurrent = objPrevious;
		m_descCurrent = descPrevious;
		toggleBufferingMode( bOldBufferingMode );
	}
		
	// If any exception occurs during the recursive calls to writeObject
	//  throw it back. If this is the base call reset m_exFatalException
	IOException exAny = m_exFatalException;

	if( m_nRecursionLevel == 0 )
	    m_exFatalException = null;
	if( exAny != null )
	    throw exAny;		
}

/**
 * Write the non-static and non-transient fields of the current class
 * to this stream.  This may only be called from the writeObject method
 * of the class being serialized. It will throw the NotActiveException
 * if it is called otherwise.
 * @since     JDK1.1
 */
public final void defaultWriteObject()
    throws IOException
{
	// Make sure that the method is called only from writeObject method
	if( (m_objCurrent == null) || (m_descCurrent == null) )
		throw new NotActiveException( "Invalid call to defaultWriteObject" );

	boolean bOldBufferingMode = toggleBufferingMode( false );

    // do the actual read
	invokeDefaultWriteObject( m_objCurrent, m_descCurrent.forClass() );
	toggleBufferingMode( bOldBufferingMode );
}

/**
 * Reset will disregard the state of any objects already written
 * to the stream.  The state is reset to be the same as a new
 * ObjectOutputStream.  The current point in the stream is marked
 * as reset so the corresponding ObjectInputStream will be reset
 * at the same point.  Objects previously written to the stream
 * will not be refered to as already being in the stream.  They
 * will be written to the stream again.
 * @since     JDK1.1
 */
public void reset()
	throws IOException
{
	if( (m_objCurrent != null) || (m_descCurrent != null) )
        throw new IOException( "Illegal call to reset" );

	// flush the buffer
	toggleBufferingMode( false );

	// Write the code for reset
	writeHeader( TC_RESET );
	setDefaultStreamValues();
	m_exFatalException = null;
}

/**
 * Closes the stream. This method must be called
 * to release any resources associated with the
 * stream.
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public void close()
	throws IOException
{
	flush();
	m_out.close();
}

///////////////DataOutput Methods are implemented here///////////////////////////////////////////

////////////// The next three methods write byte/s to the output stream /////////////////////////

/**
 * Writes a byte. This method will block until the byte is actually
 * written.
 * @param nByte the byte
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public void write( int nByte )
	throws IOException
{
    if( m_nBlockSize >= m_aBlockData.length )
    {
        // Buffer is full, empty it.
	    drain();
    }

	m_aBlockData[ m_nBlockSize++ ] = (byte)nByte;
}

/**
 * Writes an array of bytes. This method will block until the bytes
 * are actually written.
 * @param aBytes the data to be written
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public void write( byte aBytes[] )
	throws IOException
{
	write( aBytes, 0, aBytes.length );
}

/**
 * Writes a sub array of bytes.
 * @param aBytes the data to be written
 * @param nOffest the start offset in the data
 * @param nLength the number of bytes that are written
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public void write( byte aBytes[], int nOffset, int nLength )
	throws IOException
{
	if( nLength < 0 )
	    throw new IndexOutOfBoundsException();
	
	// If there is space in the buffer then write the array there. Otherwise, flush the
	//  buffer and then write the data.
	int nSpace = m_aBlockData.length - m_nBlockSize;

	if( nLength <= nSpace )
	{
		System.arraycopy( aBytes, nOffset, m_aBlockData, m_nBlockSize, nLength );
		m_nBlockSize += nLength;
	}
	else  // Not enough space
	{
		drain();

		if( m_bCurrentBufferingMode )
		{
			if( nLength <= 255 )
			{
				m_out.write( TC_BLOCKDATA );
				m_out.write( nLength );
			}
			else  // the data is a long data block record
			{
			    m_out.write( TC_BLOCKDATALONG );
				m_out.write( (nLength >> 24) & 0xFF );
				m_out.write( (nLength >> 16) & 0xFF );
				m_out.write( (nLength >>  8) & 0xFF );
				m_out.write( nLength & 0xFF );
			}
	    }

	    m_out.write( aBytes, nOffset, nLength );
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////

////DataOutputStream methods are implemented here. They are simply redirected calls.//////////////

/**
 * Writes a boolean.
 * @param bBoolVal the boolean to be written
 * @since     JDK1.1
 */
public void writeBoolean( boolean bBoolVal )
	throws IOException
{
	m_dos.writeBoolean( bBoolVal );
}

/**
 * Writes an 8 bit byte.
 * @param nByteVal the byte value to be written
 * @since     JDK1.1
 */
public void writeByte( int nByteVal )
	throws IOException
{
	m_dos.writeByte( nByteVal );
}

/**
 * Writes a 16 bit short.
 * @param nShortVal the short value to be written
 * @since     JDK1.1
 */
public void writeShort( int nShortVal )
	throws IOException
{
	m_dos.writeShort( nShortVal );
}

/**
 * Writes a 16 bit char.
 * @param nCharVal the char value to be written
 * @since     JDK1.1
 */
public void writeChar( int nCharVal )
	throws IOException
{
	m_dos.writeChar( nCharVal );
}

/**
 * Writes a 32 bit int.
 * @param nIntVal the integer value to be written
 * @since     JDK1.1
 */
public void writeInt(int nIntVal )
	throws IOException
{
	m_dos.writeInt( nIntVal );
}

/**
 * Writes a 64 bit long.
 * @param lLongVal the long value to be written
 * @since     JDK1.1
 */
public void writeLong( long lLongVal )
	throws IOException
{
	m_dos.writeLong( lLongVal );
}

/**
 * Writes a 32 bit float.
 * @param fFloatVal the float value to be written
 * @since     JDK1.1
 */
public void writeFloat( float fFloatVal )
	throws IOException
{
	m_dos.writeFloat( fFloatVal );
}

/**
 * Writes a 64 bit double.
 * @param dDoubleVal the double value to be written
 * @since     JDK1.1
 */
public void writeDouble( double dDoubleVal )
	throws IOException
{
	m_dos.writeDouble( dDoubleVal );
}

/**
 * Writes a String as a sequence of bytes.
 * @param strBytes the String of bytes to be written
 * @since     JDK1.1
 */
public void writeBytes( String strBytes )
	throws IOException
{
	m_dos.writeBytes( strBytes );
}

/**
 * Writes a String as a sequence of chars.
 * @param strChars the String of chars to be written
 * @since     JDK1.1
 */
public void writeChars( String strChars )
	throws IOException
{
	m_dos.writeChars( strChars );
}

/**
 * Writes a String in UTF format.
 * @param strUTF the String in UTF format
 * @since     JDK1.1
 */
public void writeUTF( String strUTF )
	throws IOException
{
	m_dos.writeUTF( strUTF );
}

/**
 * Flushes the stream. This will write any buffered
 * output bytes and flush through to the underlying stream.
 * @exception IOException If an I/O error has occurred.
 * @since     JDK1.1
 */
public void flush()
	throws IOException
{
	drain();
	m_out.flush();
}

//////////////////////////////////////////////////////////////////////////////////////////////////

//////////////Protected methods are defined here//////////////////////////////////////////////////

/**
 * Drain any buffered data in ObjectOutputStream.  Similar to flush
 * but does not propagate the flush to the underlaying stream.
 * @since     JDK1.1
 */
protected void drain()
	throws IOException
{
    // Empty out the buffer. If the buffering mode was on then
    //  write the header to the stream followed by the count
    //  and then the data in the buffer
	if( m_nBlockSize == 0)
	    return;

	if( m_bCurrentBufferingMode )
	{
	    m_out.write( TC_BLOCKDATA );
	    m_out.write( m_nBlockSize );
	}

	m_out.write( m_aBlockData, 0, m_nBlockSize );
	m_nBlockSize = 0;
}

/**
 * The writeStreamHeader method is provided so subclasses can
 * append or prepend their own header to the stream.
 * It writes the magic number and version to the stream.
 * @since     JDK1.1
 */
protected void writeStreamHeader()
	throws IOException
{
	writeShort( STREAM_MAGIC );
	writeShort( STREAM_VERSION );
}

/**
 * Enable the stream to do replacement of objects in the stream.
 * If the stream is a trusted class it is allowed to enable replacement.
 * Trusted classes are those classes with a classLoader equals null. <p>
 *
 * When enabled the replaceObject method is called for every object
 * being serialized.
 *
 * @exception SecurityException The classloader of this stream object is non-null.
 * @since     JDK1.1
 */
protected final boolean enableReplaceObject( boolean bEnable )
	throws SecurityException
{
	boolean bEnabledPreviously = m_bReplaceEnabled;

	if( bEnable )
	{
	    ClassLoader loader = this.getClass().getClassLoader();

	    if( loader == null )
		{
			m_bReplaceEnabled = true;

			return bEnabledPreviously;
	    }

	    throw new SecurityException( "Not trusted class" );
	}
	else
	{
	    m_bReplaceEnabled = false;
	}

	return bEnabledPreviously;
}

/** This method will allow trusted subclasses of ObjectOutputStream
 * to substitute one object for another during
 * serialization. Replacing objects is disabled until
 * enableReplaceObject is called. The enableReplaceObject method
 * checks that the stream requesting to do replacment can be
 * trusted. Every reference to serializable objects is passed to
 * replaceObject.  To insure that the private state of objects is
 * not unintentionally exposed only trusted streams may use
 * replaceObject. <p>
 *
 * When a subclass is replacing objects it must insure that either
 * a complementary substitution must be made during
 * deserialization or that the substituted object is compatible
 * with every field where the reference will be stored.  Objects
 * whose type is not a subclass of the type of the field or array
 * element abort the serialization by raising an exception and the
 * object is not be stored. <p>
 *
 * This method is called only once when each object is first encountered.
 * All subsequent references to the object will be redirected to the
 * new object. This method should return the object to be substituted or
 * the original object. <P>
 *
 * Null can be returned as the object to be substituted, but may
 * cause NullReferenceException in classes that contain references
 * to the original object since they may be expecting an object
 * instead of null.<p>
 *
 * @exception IOException Any exception thrown by the underlying
 * OutputStream.
 * @since     JDK1.1
 */
protected Object replaceObject( Object obj )
	throws IOException
{
	return obj;
}

/**
 * Subclasses may implement this method to allow class data to be stored
 * in the stream. By default this method does nothing.
 * The corresponding method in ObjectInputStream is resolveClass.
 * This method is called exactly once for each unique class in the stream.
 * The class name and signature will have already been written to the stream.
 * This method may make free use of the ObjectOutputStream to save
 * any representation of the class it deems suitable (for example,
 * the bytes of the class file).  The resolveClass method in the corresponding
 * subclass of ObjectInputStream must read and use any data or objects
 * written by annotateClass.
 * annotateClass is called only for normal classes.  Arrays are not normal classes.
 * @exception IOException Any exception thrown by the underlying OutputStream.
 * @since     JDK1.1
 */
protected void annotateClass( Class cl )
	throws IOException
{
}

//////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////Private helper methods///////////////////////////////////////////////////////////

// This method changes the buffering mode.
// If buffering is being stoppped then the buffer is drained
private boolean toggleBufferingMode( boolean bNewBufferingMode )
	throws IOException
{
	if( m_bCurrentBufferingMode == bNewBufferingMode )
		return bNewBufferingMode;

	drain();

	m_bCurrentBufferingMode = bNewBufferingMode;

    // The previous mode is returned
	return !bNewBufferingMode;
}

// reset the variables to their default state
private void setDefaultStreamValues()
	throws IOException
{
	m_bReplaceEnabled  = false;
	m_mapObjectToHandle = new OrdinalMap();
	m_htOriginalToAlternateObject = new Hashtable( 11 );
	toggleBufferingMode( true );
}



private boolean checkNullAndPreviousReference( Object obj )
	throws IOException
{
	if( obj == null )
	{
		writeHeader( TC_NULL );

		return true;
	}

    // Check if the object has been replaced.
    ReplaceObjectsClass replacementInfo =
                           (ReplaceObjectsClass)m_htOriginalToAlternateObject.get( obj );

    if( replacementInfo != null )
        obj = replacementInfo.m_obj;

    int nHandle = m_mapObjectToHandle.get( obj );

    if( nHandle >= 0 )
	{
		// The Object was already serialized, write a reference
		writeHeader( TC_REFERENCE );
		writeInt( nHandle + baseWireHandle );

		return true;
	}
	else
        return false;		// The object was not serialized and is non-null
}

private boolean checkNullAndPreviousReferenceNoReplace( Object obj )
	throws IOException
{
	if( obj == null )
	{
		writeHeader( TC_NULL );

		return true;
	}

    int nHandle = m_mapObjectToHandle.get( obj );

    if( nHandle >= 0 )
	{
		// The Object was already serialized, write a reference
		writeHeader( TC_REFERENCE );
		writeInt( nHandle + baseWireHandle );

		return true;
	}
	else
        return false;		// The object was not serialized and is non-null
}

private void writeHeader( int header )
	throws IOException
{
	writeByte( header );
}

// This method checks that if the object is a class, array,string
//  or a class descriptor then it is output immediately
private boolean checkSpecialCases( Object obj )
	throws IOException
{

	if( obj instanceof String )
	{
		writeString( (String)obj );

	    return true;
	}
	
	if( obj.getClass().isArray() )
	{
	    writeArray( obj );

	    return true;
	}

	if( obj instanceof ObjectStreamClass )
	{
		writeClassDescriptor( (ObjectStreamClass)obj );

	    return true;
	}

	if( obj instanceof Class )
	{
		writeClass( (Class)obj );

	    return true;
	}

	return false;
}

private void writeString( String str )
	throws IOException
{
    // write the string to the stream and assign it a handle
	writeHeader( TC_STRING );
	writeUTF( str );
	makeNewHandle( str );
}

private void writeClass( Class cl )
	throws IOException
{
	writeHeader( TC_CLASS );

	// Write the Class Descriptor and assign it a handle
	ObjectStreamClass descriptor = ObjectStreamClass.lookup( cl );

	if( descriptor == null )
        throw new NotSerializableException( cl.getName() );


	writeClassDescriptor( descriptor );
	makeNewHandle( cl );
}

// Write the Class Descriptor
private void writeClassDescriptor( ObjectStreamClass descriptor )
	throws IOException
{
	if( checkNullAndPreviousReferenceNoReplace( descriptor ) )
		return;

    // Write the class descriptor header followed by the
    //  class name and its serialVersionUID
    // assign a handle to the descriptor object
    writeHeader( TC_CLASSDESC );
    writeUTF( descriptor.getName() );
    writeLong( descriptor.getSerialVersionUID() );
    makeNewHandle( descriptor );

    // Write the name and types of the non static and non transient fields of the class
    descriptor.write( this );

    // Annotate class
    boolean bOldBufferingMode = toggleBufferingMode( true );

    annotateClass( descriptor.forClass() );
    toggleBufferingMode( bOldBufferingMode );
    writeHeader( TC_ENDBLOCKDATA );

    // write superclass
    writeClassDescriptor( descriptor.getSuperclass() );
}

// Output an array to the stream
private void writeArray( Object objArray )
	throws IOException, InvalidClassException
{
	ObjectStreamClass   descriptor = ObjectStreamClass.lookup( objArray.getClass() );

	if( descriptor == null )
    	throw new NotSerializableException( objArray.getClass().getName() );

	// Write out the header for an array and then write the
    //  descriptor of the class
	writeHeader( TC_ARRAY );
	writeClassDescriptor( descriptor );

	// assign a new handle for this object
	makeNewHandle( objArray );

	// write the array values
    try
    {
	    switch( descriptor.getName().charAt( 1 ) )
	    {
		    case 'I': writeIntArray( (int[])objArray );         break;
		    case 'B': writeByteArray( (byte[])objArray );       break;
		    case 'S': writeShortArray( (short[])objArray );     break;
		    case 'J': writeLongArray( (long[])objArray );       break;
		    case 'F': writeFloatArray( (float[])objArray );     break;
		    case 'D': writeDoubleArray( (double[])objArray );   break;
		    case 'C': writeCharArray( (char[])objArray );       break;
		    case 'Z': writeBooleanArray( (boolean[])objArray ); break;

		    case '[':
		    case 'L': writeObjectArray( (Object[])objArray );   break;

		    default: throw new InvalidClassException( descriptor.getName() );
	    }
    }
    catch( ClassCastException exClassCast )
    {
        throw new InvalidClassException( descriptor.getName() );
    }
}

// Write the object to the stream.
private void writeNewObject( Object obj )
	throws IOException
{
	m_objCurrent = obj;

	// Lookup the class descriptor for the class
	// If null is returned then this class doesn't implement
	//  serializable interface and an exception is thrown
	m_descCurrent = ObjectStreamClass.lookup( obj.getClass() );
	if( m_descCurrent == null )
	    throw new NotSerializableException( obj.getClass().getName() );

	// Output the header for a new object and then
	//  output the class descriptor
	writeHeader( TC_OBJECT );
	writeClassDescriptor( m_descCurrent );

	// Make a new handle for this object
	makeNewHandle( obj );

	if( m_descCurrent.isExternalizable() )
	    ((Externalizable)m_objCurrent).writeExternal( this );
	else
    {
        // hand off
        writeFieldData( m_descCurrent );
    }	
}

private void writeFieldData( ObjectStreamClass descriptor )
	throws IOException, NotSerializableException
{
    if( descriptor == null )
		return;
	else
    {
        // recursive call to write superclass data
		writeFieldData( descriptor.getSuperclass() );
    }

    // if the current object has a writeObject() method, call it.  else, perform
    //  default write.
	m_descCurrent = descriptor;
	if( m_descCurrent.hasWriteObject() )
	{
        // Buffer the data the class writes
		toggleBufferingMode( true );
		invokeWriteObject( m_objCurrent, descriptor.forClass() );
		toggleBufferingMode( false );
		writeHeader( TC_ENDBLOCKDATA );
	}
	else
		defaultWriteObject();
}

// Generate a New Handle
private void makeNewHandle( Object obj )
	throws IOException
{
	m_mapObjectToHandle.put( obj );
}

//Add the original object and alternate object to the table
private void mapOriginalToAlternate( Object objOriginal, Object objAlternate )
{
	ReplaceObjectsClass replacementInfo = new ReplaceObjectsClass( objAlternate );
	m_htOriginalToAlternateObject.put( objOriginal, replacementInfo );
}

////////////////Methods to handle arrays//////////////////////////////////////////////////////////

private void writeIntArray( int[] aInts )
	throws IOException
{
	writeInt( aInts.length );
	for( int index = 0; index < aInts.length; index++ )
		writeInt( aInts[ index ] );
}

private void writeByteArray( byte[] aBytes )
	throws IOException
{
	writeInt( aBytes.length );
	for( int index = 0; index < aBytes.length; index++ )
		writeByte( (int)aBytes[ index ] );
}
	
private void writeShortArray( short[] aShorts )
	throws IOException
{
	writeInt( aShorts.length );
	for( int index = 0; index < aShorts.length; index++ )
		writeShort( (int)aShorts[ index ] );
		
}

private void writeCharArray( char[] aChars )
	throws IOException
{
	writeInt( aChars.length );
	for( int index = 0; index < aChars.length; index++ )
		writeChar( (int)aChars[ index ] );
}

private void writeLongArray( long[] aLongs )
	throws IOException
{
	writeInt( aLongs.length );
	for( int index = 0; index < aLongs.length; index++ )
		writeLong( aLongs[ index ] );
}

private void writeFloatArray( float[] aFloats )
	throws IOException
{
	writeInt( aFloats.length );
	for( int index = 0; index < aFloats.length; index++ )
		writeFloat( aFloats[ index ] );
}

private void writeDoubleArray( double[] aDoubles )
	throws IOException
{
	writeInt( aDoubles.length );
	for( int index = 0; index < aDoubles.length; index++ )
		writeDouble( aDoubles[ index ] );
}

private void writeBooleanArray( boolean[] aBooleans )
	throws IOException
{
	writeInt( aBooleans.length );
	for( int index = 0; index < aBooleans.length; index++ )
		writeBoolean( aBooleans[ index ] );
}

private void writeObjectArray( Object[] aObjects )
	throws IOException
{
	writeInt( aObjects.length );
	for( int index = 0; index < aObjects.length; index++ )
		writeObject( aObjects[ index ] );
}

}   // end of class ObjectInputStream

//////////////////////////////////////////////////////////////////////////////////////////////////	

// Class for holding a replacement object
class ReplaceObjectsClass
{
	Object  m_obj;

	ReplaceObjectsClass( Object obj )
	{
		m_obj = obj;
	}
}
