/*
 * @(#)ObjectStreamClass.java	1.30 97/01/27
 * 
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.1_beta
 * Copyright (C) 1997 Microsoft Corporation. Portions copyright (C) 1993-1997 Sun Microsystems, Inc. All Rights Reserved.
 */

package java.io;

import java.util.Hashtable;
import java.lang.reflect.Modifier;
import java.lang.reflect.Member;
import java.lang.reflect.Field; 
import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.DigestOutputStream;
import com.ms.security.PolicyEngine;
import com.ms.security.PermissionID;
import com.ms.vm.WeakReference;

/**
 * A ObjectStreamClass describes a class that can be serialized to a stream
 * or a class that was serialized to a stream.  It contains the name
 * and the serialVersionUID of the class.
 * <br>
 * The ObjectStreamClass for a specific class loaded in this Java VM can
 * be found using the lookup method.
 *
 * @since   JDK1.1
 */
public class ObjectStreamClass  
	implements java.io.Serializable 
{ 
  	// Class this descriptor represents 
	private Class               m_classRep;

  	// Version and name of this class
	private long                m_lSerialVersionUID;
	private String              m_strClassName;

    // descriptor of super class
	private ObjectStreamClass   m_descSuperClass;

	// Stores the descriptors 
	private static ObjectStreamClassHashtable    m_htDescCache = new ObjectStreamClassHashtable();
	
    // Names and types of fields read or written, as well as the
    //  number of such fields
	private String[]            m_aFieldNames;
	private byte[]              m_aFieldTypeByte;
	private int                 m_nFields;

    // Field object for each valid field of the local class
    private Field[]             m_aFields;

	// Flags to tell whether custom read and write methods are present or not.
	private boolean             m_bWriteMethodPresent;
	private boolean             m_bReadMethodPresent;

	// Flag to tell whether the versions match
	private boolean             m_bVersionMismatch;

	// flags to tell which interface is present 
	private boolean             m_bSerializable;
	private boolean             m_bExternalizable;

    // serializable (java.io.Serializable) and externalizable
    //  (java.io.Externalizable) classes
	private static Class        m_classSerializable;
	private static Class        m_classExternalizable;

    // get the serialVersionUID from the specified class
    private static native long getClassDefinedUID( Class cl );

    // Read in the list of fields that are marked non-null in the m_aFields member
    // of the ObjectStreamClass param.
    // This method is called only when the versions reading and writing the streams
    //  are different.
    native void doMismatchedRead( ObjectInputStream ois, Object obj )
	    throws IOException, ClassNotFoundException;

	private static final int METHOD_TYPE_CLINIT = 0;
	private static final int METHOD_TYPE_READ = 1;
	private static final int METHOD_TYPE_WRITE = 2;

    // returns true if the class has a class/static initializer, else false
	native boolean findObjectMethod0( Class cl, int type );

///////////////////////Static initializers////////////////////////////////////////////////////////

	//Initialize the serializable and externalizable classes upfront 
	static
	{
		try
		{
			m_classSerializable = Class.forName( "java.io.Serializable" );
			m_classExternalizable = Class.forName( "java.io.Externalizable" );
		}
		catch( ClassNotFoundException exCNF )
		{
			// neutralize: no class will be serializable
	        System.err.println( "Could not load java.io.Serializable or java.io.Externalizable." );
		}
	}

//////////////////////////////////////////////////////////////////////////////////////////////////

// Constructor called when the descriptor is not found in the cache and the class is serializable
private ObjectStreamClass( Class cl, boolean bSerializable, boolean bExternalizable )
	throws IOException, SecurityException, NotSerializableException
{
	m_classRep  = cl;
	m_strClassName = cl.getName();
	m_bSerializable = bSerializable;
	m_bExternalizable = bExternalizable;

	// Generate a serial version UID
	m_lSerialVersionUID = generateSerialVersionUID( cl );

	if( m_bSerializable )
	{	
		// Use reflection to find whether the class has it's own read/writeObjectMethod
		findWriteObjectMethod();
		findReadObjectMethod();
	
		Class   classSuper = cl.getSuperclass();
		if( classSuper != null )
		    m_descSuperClass = lookup( classSuper );
		else
            m_descSuperClass = null;
	}
}

// Constructor called when a descriptor is first encountered in ObjectInputStream
ObjectStreamClass( String strClassName, long lSerialVersion )
{
	m_strClassName = strClassName;
	m_lSerialVersionUID = lSerialVersion;
}

////////////////////Public methods////////////////////////////////////////////////////////////////

/** Find the descriptor for a class that can be serialized.  Null
 * is returned if the specified class does not implement
 * java.io.Serializable or java.io.Externalizable.
 * @since   JDK1.1
 */
public static synchronized ObjectStreamClass lookup( Class cl ) 
{
	// Have we seen this class before?
    ObjectStreamClass   desc = m_htDescCache.get( cl );

	if( desc != null )
    {
	    return desc;
    }

	// If the class implements the serializable interface then 
	//  return a new ObjectStream Descriptor else return null 	

    // check to make sure class implements serializable or externalizable
	boolean bSerializable = m_classSerializable.isAssignableFrom( cl );

	if( !bSerializable )
		return null;

	boolean bExternalizable = m_classExternalizable.isAssignableFrom( cl );

	if( bExternalizable )
		bSerializable = false;

	try
	{
		// create the descriptor and add it to the cache
		desc = new ObjectStreamClass( cl, bSerializable, bExternalizable );

		m_htDescCache.put( cl, new WeakReference( desc ) );

		return desc;
	}
	catch( IOException e ) 
	{
		return null;
	}
}

/**
 * The name of the class described by this descriptor.
 * @since   JDK1.1
 */
public String getName()
{
	return m_strClassName;
}

/**
 * Return the class in the local VM that this version is mapped to.
 * Null is returned if there is no corresponding local class.
 * @since   JDK1.1
 */
public Class forClass()
{
	return m_classRep;
}

/**
 * Return the serialVersionUID for this class.
 * The serialVersionUID defines a set of classes all with the same name
 * that have evolved from a common root class and agree to be serialized
 * and deserialized using a common format.
 * @since   JDK1.1
 */
public long getSerialVersionUID()
{
	return m_lSerialVersionUID;
}

/**
 * Return a string describing this ObjectStreamClass.
 * @since   JDK1.1
 */
public String toString()
{
	return m_classRep.toString();
}

//////////////////////////////////////////////////////////////////////////////////////////////////


/////////////Package accessible methods///////////////////////////////////////////////////////////
    
// return the superclass of the class that this descriptor represents 
ObjectStreamClass getSuperclass() 
{ 
	return m_descSuperClass;
}

// flag to indicate that the local class and the class in the stream have different versions
boolean isVersionMismatched() 
{
	return m_bVersionMismatch;
}

boolean hasWriteObject() 
{
	return m_bWriteMethodPresent;
}

boolean hasReadObject()
{
	return m_bReadMethodPresent;
}

boolean isSerializable() 
{
	return m_bSerializable;
}

boolean isExternalizable()
{
	return m_bExternalizable;
}

boolean typeEquals( ObjectStreamClass other )
{
	return( (m_lSerialVersionUID == other.m_lSerialVersionUID) &&
                           m_strClassName.equals( other.m_strClassName ) );
}

void write( ObjectOutputStream oos )
   throws IOException
{
    byte    flag = 0;

    // write flags
	if( m_bSerializable )
		flag |= ObjectStreamConstants.SC_SERIALIZABLE;
	if( m_bExternalizable )
		flag |= ObjectStreamConstants.SC_EXTERNALIZABLE;
	if( m_bWriteMethodPresent )
		flag |= ObjectStreamConstants.SC_WRITE_METHOD;

	oos.writeByte( flag );

    if( m_bExternalizable )
    {
        // no fields written for class that implement Externalizable, per JDK implementation
        oos.writeShort( 0 );
    }
    else
    {
	    // Write the non static and non transient fields of the class.
        oos.writeShort( m_nFields );

	    for( int index = 0; index < m_nFields; index++ )
	    { 
		    Class   classFieldType = m_aFields[ index ].getType();
		    byte    fieldType = (byte)getFieldType( classFieldType );

		    oos.writeByte( fieldType );
		    oos.writeUTF( m_aFields[ index ].getName() );

		    if( fieldType == 'L' )
			    oos.writeObject( "L" + classFieldType.getName().replace( '.', '/' ) + ";" );
		    else if( fieldType == '[' )
			    oos.writeObject( classFieldType.getName().replace( '.', '/' ) );
	    }
    }
}

// Read in the number of fields, their names and types. If there is a
//  version mismatch, find the subset of fields to be read. This
//  information is gathered only once per class, when the descriptor
//  for the class is first encountered in the stream.
void read( ObjectInputStream ois )
   throws IOException, ClassNotFoundException
{ 
	// Read in the byte flag and set our internal values accordingly
	byte    flag = ois.readByte();

	m_bWriteMethodPresent = ((flag & ObjectStreamConstants.SC_WRITE_METHOD) != 0);
	m_bSerializable = ((flag & ObjectStreamConstants.SC_SERIALIZABLE) != 0);
	m_bExternalizable = ((flag & ObjectStreamConstants.SC_EXTERNALIZABLE) != 0);

    // read the fields
	m_nFields = ois.readShort();
	if( m_nFields < 0 )
		throw new StreamCorruptedException();

	if( m_nFields != 0 )
	{
		m_aFieldTypeByte = new byte[ m_nFields ];
		m_aFieldNames = new String[ m_nFields ];

		for( int index = 0; index < m_nFields; index++ ) 
		{	 
			m_aFieldTypeByte[ index ] = ois.readByte();
			m_aFieldNames[ index ] = ois.readUTF();

            // read type string if need be
			if( (m_aFieldTypeByte[ index ] == 'L') || (m_aFieldTypeByte[ index ] == '[') )
            {
                // read the class name.  we never actually use this info.
		        if( !(ois.readObject() instanceof String ) )
		            throw new StreamCorruptedException();
            }

		}	
	}
}

void setSuperClass( ObjectStreamClass descSuperClass )
{
	m_descSuperClass = descSuperClass;
}

void setClass( Class cl )
    throws InvalidClassException
{
    // try to find a local version of this class
	ObjectStreamClass   descLocal = lookup( cl );

	if( !descLocal.getName().equals( m_strClassName ) ||
		    (descLocal.m_lSerialVersionUID != m_lSerialVersionUID) )
	{
	    throw new InvalidClassException( cl.getName(), "Local class not compatible" );
	}

	if( (m_bSerializable != descLocal.isSerializable()) ||
		        (m_bExternalizable != descLocal.isExternalizable()) )
	    throw new InvalidClassException( cl.getName(),
					        "Serialization incompatible with Externalization" );

	m_classRep = cl;

	// At this point the class is known, so we can use information in the
    //  class' descriptor to find out which methods it supports.
	if( !m_bReadMethodPresent && !cl.isArray() ) 
		findReadObjectMethod();

	if( m_nFields != 0 )
	{
		// Now check the list of fields of local class against the list of
        //  fields just read in and find the subset of fields to be read.
        Field[]         aLocalFields = descLocal.m_aFields;
        int             nNumLocalFields = descLocal.m_nFields;

		m_aFields = new Field[ m_nFields ];     // null element => don't read

		int nLocalFieldIndex = 0;
		int nNumFieldsFoundLocally = 0;

		for( int nStreamFieldIndex = 0; nStreamFieldIndex < m_nFields; nStreamFieldIndex++ )
		{
            m_aFields[ nStreamFieldIndex ] = null;

			for( ; nLocalFieldIndex < nNumLocalFields; nLocalFieldIndex++ )
			{
                // check if this local fields matches the field read from the stream
				if( aLocalFields[ nLocalFieldIndex ].getName().equals(
                                            m_aFieldNames[ nStreamFieldIndex ] ) )
				{
                    if( getFieldType( aLocalFields[ nLocalFieldIndex ].getType() ) !=
                                                    m_aFieldTypeByte[ nStreamFieldIndex ] )
						throw new InvalidClassException(
                                            "Type change for field " +
											m_aFieldNames[ nStreamFieldIndex ] + 
											" has made the class incompatible" );

					m_aFields[ nStreamFieldIndex ] = aLocalFields[ nLocalFieldIndex ];
					nNumFieldsFoundLocally++;
					break;
				}
			}
 
            // Next position to start searching from
			nLocalFieldIndex++;
		}

		// A version mismatch occurs when the number of fields in the stream
        //  and the local class are not the same or when all the fields that
        //  are present in the stream are not present in the local class.
        if( (m_nFields != nNumLocalFields) || (nNumFieldsFoundLocally != m_nFields) )
		{
			m_bVersionMismatch = true;
		}
        else
        {
			m_bVersionMismatch = false;
        }

        // null out unused array space
        m_aFieldNames = null;
	}
}

// Skip the data for classes that are present in the stream but not in the
//  local class hierarchy
void skipFieldData( ObjectInputStream ois ) 
	throws IOException, StreamCorruptedException, ClassNotFoundException
{
	for( int index = 0; index < m_nFields; index++ ) 
	{
		switch( (char) m_aFieldTypeByte[ index ] ) 
		{
			case 'I':   ois.readInt();      break;
			case 'L':
            case '[':   ois.readObject();   break;
			case 'C':   ois.readChar();     break;
			case 'Z':   ois.readBoolean();	break;
			case 'F':   ois.readFloat();    break;
			case 'S':   ois.readShort();	break;
			case 'B':   ois.readByte();		break;
			case 'J':   ois.readLong();		break;
			case 'D':   ois.readDouble();	break;

			default:
                throw new StreamCorruptedException( "Unknown field type in stream for " +
															                    m_classRep );
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
 
////////////////////Private helper methods////////////////////////////////////////////////////////

// Use reflection to determine if the read/writeObject method exists
private void findWriteObjectMethod() 
{    
	m_bWriteMethodPresent = findObjectMethod0(m_classRep, METHOD_TYPE_WRITE);
}

// Use reflection to determine if the readObject method exists 
private void findReadObjectMethod() 
{
	m_bReadMethodPresent = findObjectMethod0(m_classRep, METHOD_TYPE_READ);
}

// The class doesn't have it's own uid defined. Generate it using sha-1 algorithm.
//  This functions also creates the m_aFields array.
private long generateSerialVersionUID( Class cl )
{
	long    lUID = getClassDefinedUID( cl );

	if( lUID != 0 )
    {
        // before returning this ID, we need to create the field array.
        //  use reflection to get info about this class.
        PolicyEngine.assertPermission(PermissionID.REFLECTION);
	    Field[]         aClassFields = cl.getDeclaredFields();
        PolicyEngine.revertPermission(PermissionID.REFLECTION);

        m_aFields = new Field[ aClassFields.length ];

        // we don't want fields that are transient or static
        m_nFields = 0;
	    for( int index = 0; index < aClassFields.length; index++ ) 
		{			
			int fieldModifier = aClassFields[ index ].getModifiers();

            if( ((fieldModifier & Modifier.TRANSIENT) == Modifier.TRANSIENT) ||
				                    ((fieldModifier & Modifier.STATIC) == Modifier.STATIC) )
            {
                continue;
            }

            m_aFields[ m_nFields++ ] = aClassFields[ index ];
	    }

	    // Sort the field names
	    FieldArraySort fas = new FieldArraySort( m_aFields );

		fas.sort( m_nFields );

        // return the UID
		return lUID;
    }

	try 
	{
		BlackHoleOutputStream   bos = new BlackHoleOutputStream();
		MessageDigest           md = MessageDigest.getInstance( "SHA" );
	    DigestOutputStream      dgo = new DigestOutputStream( bos, md );
	    DataOutputStream        dos = new DataOutputStream( dgo );

        // use reflection to get info about this class
        PolicyEngine.assertPermission(PermissionID.REFLECTION);
	    Field[]         aClassFields = cl.getDeclaredFields();
		Method[]        aMethods = cl.getDeclaredMethods();
		Constructor[]   aConstructors = cl.getDeclaredConstructors();
        PolicyEngine.revertPermission(PermissionID.REFLECTION);

		// Class name written using UTF encoding 
		dos.writeUTF( cl.getName() );

		// Class modifiers written as a 32-bit integer.  as per Sun
        //  implementation, mask the modifiers.
		int classModifier = cl.getModifiers();

		classModifier &= (Modifier.INTERFACE | Modifier.ABSTRACT |
                                        Modifier.FINAL | Modifier.PUBLIC);
		dos.writeInt( classModifier );

		// Write the names of the interfaces in lexical order
		Class[]         aInterfaces = cl.getInterfaces();
		ClassArraySort  cas = new ClassArraySort( aInterfaces );

	    cas.sort( aInterfaces.length );
		
	    for( int index = 0; index < aInterfaces.length; index++ ) 
		{
            dos.writeUTF( aInterfaces[ index ].getName() );
		}

	    // Sort the field names
	    MemberArraySort mas = new MemberArraySort( aClassFields );

		mas.sort( aClassFields.length );
	    
	    // Write all the fields except those that are private static or private transient.
        // while we're at it, remember those fields that will be serialized (i.e., those
        //  that are neither static nor transient.
	    m_aFields = new Field[ aClassFields.length ];
        m_nFields = 0;
	    for( int index = 0; index < aClassFields.length; index++ ) 
		{			
			int fieldModifier = aClassFields[ index ].getModifiers();

            if( ((fieldModifier & Modifier.TRANSIENT) == Modifier.TRANSIENT) ||
				                    ((fieldModifier & Modifier.STATIC) == Modifier.STATIC) )
            {
                if( (fieldModifier & Modifier.PRIVATE) == Modifier.PRIVATE )
                    continue;
            }
            else
            {
                m_aFields[ m_nFields++ ] = aClassFields[ index ];
            }

			// Write the name, modifiers, and signature
			dos.writeUTF( aClassFields[ index ].getName() );
			dos.writeInt( fieldModifier );
			dos.writeUTF( getTypeSignature(
                                aClassFields[ index ].getType() ).replace( '.', '/' ) );
	    }

        // now sort the fields in the order in which they must be written to the stream
        FieldArraySort  fas = new FieldArraySort( m_aFields );

        fas.sort( m_nFields );
	    
        // get all methods, including ctors and class initializer
        MethodInfo[]    aMethodInfo = new MethodInfo[ aMethods.length +
                                                            aConstructors.length + 1 ];
        int             nMethodIndex = 0;
		
		for( int index = 0; index < aMethods.length; index++ )
        {
            Method  method = aMethods[ index ];

			if( (method.getModifiers() & Modifier.PRIVATE) != 0 )
				continue;

            aMethodInfo[ nMethodIndex++ ] = new MethodInfo( method.getName(),
                                                            createMethodSignature(
                                                                method.getParameterTypes(),
                                                                method.getReturnType() ),
                                                            method.getModifiers() );
        }

		for( int index = 0; index < aConstructors.length; index++ )
        {
            Constructor ctor = aConstructors[ index ];

			if( (ctor.getModifiers() & Modifier.PRIVATE) != 0 )
				continue;
                        
            aMethodInfo[ nMethodIndex++ ] = new MethodInfo( "<init>",
                                                            createMethodSignature(
                                                                ctor.getParameterTypes(),
                                                                Void.TYPE ),
                                                            ctor.getModifiers() );
        }

		if( findObjectMethod0( cl, METHOD_TYPE_CLINIT ) )
            aMethodInfo[ nMethodIndex++ ] = new MethodInfo( "<clinit>",
                                                                "()V",
                                                                Modifier.STATIC );

		// Sort the method info
		MethodInfoArraySort sas = new MethodInfoArraySort( aMethodInfo );

		sas.sort( nMethodIndex );
	    
        // write each method's name, modifiers, and signature to the string
	    for( int index = 0; index < nMethodIndex; index++ )
		{
			dos.writeUTF( aMethodInfo[ index ].m_strName );
			dos.writeInt( aMethodInfo[ index ].m_mods );
			dos.writeUTF( aMethodInfo[ index ].m_strSignature );
	    }

		// Note the ordering of the bytes. This follows SUN's implementation but
        //  is not the same as what the specs say.
		byte    result[] = md.digest();

	    for( int index = 0; index < Math.min( 8, result.length ); index++ ) 
		{
			lUID += (long)(result[ index ] & 255) << (index * 8);
	    }

		/*int result[] = algo.getResults();
		uid = (((long)result[1]) << 32) | (((long)result[0]) & 0xFFFFFFFF);*/
	}
	catch( NoSuchAlgorithmException exNoSuchAlgorithm )
	{
	    throw new SecurityException( exNoSuchAlgorithm.getMessage() );
	}
	catch( IOException e )
	{
		lUID = -1;    
	}
	return lUID;
}

private String getTypeSignature( Class classType )
{
	if( classType == Void.TYPE )
		return "V";
	else if( classType.isPrimitive() )
		return( String.valueOf( getFieldType( classType ) ) );
	else if( classType.isArray() )
		return classType.getName();
	else
        return "L" + classType.getName() + ";";
}

private String createMethodSignature( Class[] aParamClasses, Class classRetType )
{
    String  strSignature = "(";

    for( int index = 0; index < aParamClasses.length; index++ )
    {
        strSignature += getTypeSignature( aParamClasses[ index ] );
    }

    strSignature += ")";

    strSignature += getTypeSignature( classRetType );

    return strSignature;
}

private char getFieldType( Class classType )
{
	if( classType.isPrimitive() ) 
	{
        if( classType == Integer.TYPE )
			return 'I';
		else if( classType == Boolean.TYPE ) 
			return 'Z';
		else if( classType == Character.TYPE ) 
			return 'C'; 
		else if( classType == Float.TYPE ) 
			return 'F';
		else if( classType == Byte.TYPE ) 
			return 'B'; 
		else if( classType == Short.TYPE ) 
			return 'S';
		else if( classType == Long.TYPE ) 
			return 'J';
		else if( classType == Double.TYPE ) 
			return 'D';
	}
	else if( classType.isArray() )
		return '[';

	return 'L';
}

private void setSerialVersionUID( long lUID )
{
	m_lSerialVersionUID = lUID;
}


///////////////////////Inner classes//////////////////////////////////////////////////////////////

//
// ClassArraySort
//
private static class ClassArraySort extends com.ms.util.Sort
{
	Class[] m_aClass;

	public void sort( int high)
	{
        doSort( 0, high );
	}

	public	ClassArraySort( Class[] aClass )
	{
		m_aClass = aClass;
	}

	public void swap (int i, int j)
	{
		Class	temp = m_aClass[ i ];

		m_aClass[ i ] = m_aClass[ j ];
		m_aClass[ j ] = temp;
	}

    // simple comparison by name
	public int compare( int i, int j )
	{
		return m_aClass[ i ].getName().compareTo( m_aClass[ j ].getName() );
		
	}
}


//
// FieldArraySort
//
private static class FieldArraySort extends com.ms.util.Sort
{
	Field[] m_aFields;

	public void sort( int high )
	{
		doSort( 0, high );
	}

	public	FieldArraySort( Field[] aFields )
	{
		m_aFields = aFields;
	}

	public void swap( int i, int j )
	{
		Field	temp = m_aFields[ i ];

		m_aFields[ i ] = m_aFields[ j ];
		m_aFields[ j ] = temp;
	}

    // Serialization of fields is ordered by:
    //		primitive fields precede object references
    //		then, ascending order of name by String.compareTo
	public int compare( int i, int j )
	{
		Field	field1 = m_aFields[ i ];
		Field	field2 = m_aFields[ j ];

		boolean bIsFirstPrim = field1.getType().isPrimitive();
		boolean bIsSecondPrim = field2.getType().isPrimitive();
		if( bIsFirstPrim != bIsSecondPrim )
		{
            return( bIsFirstPrim ? -1 : 1 );
		}
		else
		{
			return field1.getName().compareTo( field2.getName() );
		}
	}
}

//
// MemberArraySort
//
private static class MemberArraySort extends com.ms.util.Sort
{
	Member[]	m_aMember;

	public void sort( int high )
	{
		doSort( 0, high );
	}

	public	MemberArraySort( Member[] aMember )
	{
		m_aMember = aMember;
	}

	public void swap( int i, int j )
	{
		Member	temp = m_aMember[ i ];

		m_aMember[ i ] = m_aMember[ j ];
		m_aMember[ j ] = temp;
	}

    // simple comparison by name
	public int compare( int i, int j )
	{
		return m_aMember[ i ].getName().compareTo( m_aMember[ j ].getName() );		
	}
}

//
// MethodInfo
//
private static class MethodInfo
{
    String  m_strName;
    String  m_strSignature;
    int     m_mods;

    MethodInfo( String strName, String strSignature, int mods )
    {
        m_strName = strName;
        m_strSignature = strSignature;
        m_mods = mods;
    }
}

//
// MethodInfoArraySort
//
static private class MethodInfoArraySort extends com.ms.util.Sort
{
	MethodInfo[]	m_aMethodInfo;

	public void sort( int high )
	{
		doSort( 0, high );
	}

	public	MethodInfoArraySort( MethodInfo[] aMethodInfo )
	{
		m_aMethodInfo = aMethodInfo;
	}

	public void swap( int i, int j )
	{
		MethodInfo	temp = m_aMethodInfo[ i ];

		m_aMethodInfo[ i ] = m_aMethodInfo[ j ];
		m_aMethodInfo[ j ] = temp;
	}

	public int compare( int i, int j )
	{
        int nCompareVal = m_aMethodInfo[ i ].m_strName.compareTo(
                                        m_aMethodInfo[ j ].m_strName );

        if( nCompareVal == 0 )
            return( m_aMethodInfo[ i ].m_strSignature.compareTo(
                                           m_aMethodInfo[ j ].m_strSignature ) );
        else
            return( nCompareVal );
	}
}

//
// BlackHoleOutputStream - eats any bytes asked to be written
//
private static class BlackHoleOutputStream extends OutputStream 
{
    // just eat the byte
	public void write( int b )
		throws IOException
	{
	}

    // just eat the bytes
    public void write( byte[] b )
        throws IOException
    {
    }

    // just eat the bytes
    public void write( byte[] b, int off, int len )
        throws IOException
    {
    }
}

//
// hashtable for ObjectStreamClass objects
//
private static class ObjectStreamClassHashtable
{
    private final static int  primes[] = {
        11,23,37,59,89,131,197,293,431,631,919,
        1327,1931,2801,4049,5839,8419,12143,
        17519,25229,36353,52361,75431, 108631, 
        156437, 225307, 324449, 467237, 672827,
        968897, 1395263, 2009191, 2893249, 4166287, 
        5999471, 7199369 
    };
  
    
    // The hash table data.
    private WeakReference   values[];
    private int             hashes[];

    // Deleted entries have their key set to deleted
    private static WeakReference deleted = new WeakReference( primes );
    
    // The total number of entries in the hash table.
    private int count;

    // hash table size and size index
    private int hashsize;
    private int primeIndex;

    public ObjectStreamClassHashtable()
    {
        this.primeIndex = 1;
        this.hashsize = primes[primeIndex];
        values = new WeakReference[hashsize*4];
        hashes = new int[hashsize*4];
    }

    public ObjectStreamClass get( Class key )
    {
        // Make sure the key is not null
        if (key == null) {
	        throw new NullPointerException();
        }
        int  hashcode = key.hashCode()& 0x7FFFFFFF;
        int  seed = hashcode;
        int  ntry = 0;

        do
        {
            int bucketNumber = 4 * (seed % hashsize);
            for (int i = 0; i < 4; i++) {
                if( (hashes[bucketNumber +i] & 0x7FFFFFFF) == hashcode ) {                    
                    WeakReference value = values[bucketNumber+i];

                    if (value == null)
                        return null;

                    if( value == deleted )
                        continue;

                    Object   ref = value.getReferent();

                    if( ref != null )
                    {
                        if( key.equals( ((ObjectStreamClass)ref).forClass() ) )
                            return ((ObjectStreamClass)ref);
                    }
                    else
                    {
                        remEntry( bucketNumber, i );
                        i--;
                    }
                }
            }

            seed += ((seed >> 5) + 1);
            seed &= 0x7FFFFFFF;
        } while (++ntry < hashsize);

        return null;
    }

    public void put( Class key, WeakReference value )
    {
        // Make sure the value is not null
        if( value == null ) {
	        throw new NullPointerException();
        }
        int     hashcode = key.hashCode()& 0x7FFFFFFF;
        int     seed = hashcode;
        int     ntry = 0;

        do {
            int bucketNumber = 4 * (seed % hashsize);

            for (int i = 0; i < 4; i++) {
                WeakReference valueCurrent = values[bucketNumber + i];

                if ((valueCurrent == null) || (valueCurrent == deleted)) {
                    values[bucketNumber + i] = value;
                    hashes [bucketNumber + i] =	(hashes [bucketNumber + i] & 0x80000000) | hashcode;
                    count++;
                    return;
                }
                if( (hashes[bucketNumber + i] & 0x7FFFFFFF) == hashcode ) {
                    Object  ref = valueCurrent.getReferent();

                    if( (ref == null) || key.equals( ((ObjectStreamClass)ref).forClass() ) ) {
                        values[bucketNumber + i] = value;
                        return;
                    }
                }
            }

            hashes[bucketNumber] |= 0x80000000;
            seed += ((seed >> 5) + 1);
            seed &= 0x7FFFFFFF;
        } while ((++ntry < 8) || (count < hashsize));

        // We need to expand to keep lookup short
        expand();

        // Try again
        put( key, value );
    }

    private void expand()
    {
        // Allocate new Array 
        int oldHashsize = hashsize;
        hashsize = primes[++primeIndex];
        WeakReference oldValues[] = values;
        values = new WeakReference[4*hashsize];
        hashes = new int [4*hashsize];

        // rehash table into it
        int nLastBucketNumber = 4 * oldHashsize;
        for( int bucketNumber = 0; bucketNumber < nLastBucketNumber; bucketNumber += 4) {
            for (int i = 0; i < 4; i++) {
                WeakReference  value = oldValues[bucketNumber+i];

                if( value == null )
                    break;

                if( value != deleted ) {
                    Object  ref = value.getReferent();

                    if( ref != null )
                        putEntry( ((ObjectStreamClass)ref).forClass(), value );
                }
            }
        }
    }

    private void putEntry( Class key, WeakReference value )
    {
        int     hashcode = key.hashCode()& 0x7FFFFFFF;
        int     seed = hashcode;
        int     ntry = 0;

        do {
            int bucketNumber = 4 * (seed % hashsize);
            for (int i = 0; i < 4; i++) {
                WeakReference   valueCurrent = values[bucketNumber+i];

                if( valueCurrent == null ) {
                    values[bucketNumber + i] = value;
                    hashes [bucketNumber + i] = (hashes [bucketNumber + i] & 0x80000000) | hashcode;	
                    return;
                }
            }

            hashes[bucketNumber] |= 0x80000000;
            seed += ((seed >> 5) + 1);
            seed &= 0x7FFFFFFF;
        } while (true);
    }

    private void remEntry (int bucketNumber, int i)
    {
        if (hashes[bucketNumber]<0) {
            values[bucketNumber + i] = deleted;
        } else {
            for (; i < 3; i++) {
                if (values[bucketNumber + i] != null) {
                    values[bucketNumber + i] = values[bucketNumber + i + 1];
                    hashes [bucketNumber+i] = hashes[bucketNumber+i+1];
                } else
                    break;
            }

            values[bucketNumber+i] = null;
        }

        count--;
    }
}

}   // end of class ObjectStreamClass
