/*
 * @(#)InetAddress.java	1.42 97/02/23
 * 
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.1_beta
 * Copyright (C) 1997 Microsoft Corporation. Portions copyright (C) 1993-1997 Sun Microsystems, Inc. All Rights Reserved.
 */

package java.net;

import java.util.Hashtable;
import com.ms.security.PolicyEngine;
import com.ms.security.PermissionID;
import com.ms.applet.BrowserAppletFrame;

/**
 * This class represents an Internet Protocol (IP) address. 
 * <p>
 * Applications should use the methods <code>getLocalHost</code>, 
 * <code>getByName</code>, or <code>getAllByName</code> to 
 * create a new <code>InetAddress</code> instance. 
 *
 * @author  Chris Warth
 * @version 1.42, 02/23/97
 * @see     java.net.InetAddress#getAllByName(java.lang.String)
 * @see     java.net.InetAddress#getByName(java.lang.String)
 * @see     java.net.InetAddress#getLocalHost()
 * @since   JDK1.0
 */
public final 
class InetAddress implements java.io.Serializable {
    String hostName;
    int address;    // Currently we only deal effectively with 32-bit addresses. 
		    // However this field can be expanded to be a byte array 
		    // or a 64-bit quantity without too much effort.
    int family;

    /** use serialVersionUID from JDK 1.0.2 for interoperability */
    private static final long serialVersionUID = 3286316764910316507L;

    /** 
     * Constructor for the Socket.accept() method.
     * This creates an empty InetAddress, which is filled in by
     * the accept() method.  This InetAddress, however, is not
     * put in the address cache, since it is not created by name.
     */
    InetAddress() {
	this.family = getInetFamily();
	}

    /**
     * Creates an InetAddress with the specified host name and IP address.
     * @param hostName the specified host name
     * @param addr the specified IP address.  The address is expected in 
     *	      network byte order.
     * @exception UnknownHostException If the address is unknown.
     */
    InetAddress(String hostName, byte addr[]) {
	this.hostName = new String(hostName);
	this.family = getInetFamily();
	/*
	 * We must be careful here to maintain the network byte
	 * order of the address.  As it comes in, the most
	 * significant byte of the address is in addr[0].  It
	 * actually doesn't matter what order they end up in the
	 * array, as long as it is documented and consistent.
	 */
	address  = addr[3] & 0xFF;
	address |= ((addr[2] << 8) & 0xFF00);
	address |= ((addr[1] << 16) & 0xFF0000);
	address |= ((addr[0] << 24) & 0xFF000000);
    }

    /**
     * Utility routine to check if the InetAddress is a 
     * IP multicast address. IP multicast address is a Class D
     * address i.e first four bits of the address are 1110.
     * @since   JDK1.1
     */
    public boolean isMulticastAddress() {
	return ((address & 0xf0000000) == 0xe0000000);
    }

    /**
     * Returns the hostname for this address.
     * If the host is equal to null, then this address refers to any
     * of the local machine's available network addresses.
     *
     * @return  the host name for this IP address.
     * @since   JDK1.0
     */
    public String getHostName()
    {
        if (hostName != null)
            return hostName;

	    String inName = null;
	    String hostaddr = getHostAddress();

        if (debug) System.out.println("finding hostname for ip "+hostaddr);

        synchronized (addressCache)
        {
            // Check if the address is already hashed.
		    Object obj = addressCache.get(hostaddr);
		    if (obj != null)
		    {
		        // Was a previous reverse-lookup marked as a permanent failure?
		        if (obj != unknown_array)
    		        inName = ((InetAddress[])obj)[0].hostName;
		    }
		    else
		    {
		        // If offline, don't touch winsock and don't fill
		        // in the hostname - we may be able to resolve it later.
                if (BrowserAppletFrame.checkOfflineContext())
                {
                    if (debug) System.out.println("offline, can't get hostname for "+this);
                    return hostaddr;
                }
                
        	    try
        	    {
        		    Object[] lookup = lookupHostByAddr(address);
        		    String name = (String)lookup[0];
        		    obj = addressCache.get(name);
        		    if (obj != null)
        		    {
        		        // Verify that the address was reported in the previous
        		        // lookup.
        		        if (obj == unknown_array)
        		        {
        		            // The first DNS lookup failed, do nothing.
        		        }
        		        else
        		        {
        		            // Check that the requested address is in the cached
        		            // hostname's list of addresses.
        		            InetAddress[] addrs = (InetAddress[])obj;
        		            for (int i = 0; ; )
        		            {
        		                if (addrs[i].address == address)
        		                {
        		                    inName = name;
        		                    break;
        		                }
        		                i++;
        		                if (i >= addrs.length)
        		                {
        		                    // Not in the list, do nothing.
        		                    break;
        		                }
        		            }
        		        }
        		    }
        		    else
        		    {
        		        // Update the cache with the information and grab
        		        // the cached InetAddress's name.
        		        inName = updateAddressCache(lookup)[0].hostName;
        		    }
        	    }
        	    catch (UnknownHostException e)
        	    {
        	        // Not offline.  Cache this failure permanently.  Note that
        	        // if running under the winsock proxy client, this will
        	        // be a common case.  If the address is discovered while
        	        // resolving a hostname, this will be replaced with the
        	        // real address.  Inconsistent, but under wspc this is the
        	        // *only* way to get reverse lookups to work at all.
                    addressCache.put(hostaddr, unknown_array);
        	    }
            }
		}

        // Don't set the hostName, the user may have disconnected from the
        // net and we may be able to resolve it later.  Also, there's no
        // point in doing a security check on the numeric address we return
        // since a check wasn't done to begin with to create this object.

        if (inName == null)
            return hostaddr;
            
        SecurityManager security = System.getSecurityManager();
        if (security != null && !security.getInCheck()) {
            security.checkConnect(inName, -1);
        }

        hostName = inName;

        return hostName;
    }

    /**
     * Returns the raw IP address of this <code>InetAddress</code> 
     * object. The result is in network byte order: the highest order 
     * byte of the address is in <code>getAddress()[0]</code>. 
     *
     * @return  the raw IP address of this object.
     * @since   JDK1.0
     */
    public byte[] getAddress() {	
	byte[] addr = new byte[4];

	addr[0] = (byte) ((address >>> 24) & 0xFF);
	addr[1] = (byte) ((address >>> 16) & 0xFF);
	addr[2] = (byte) ((address >>> 8) & 0xFF);
	addr[3] = (byte) (address & 0xFF);
	return addr;
    }

    /**
     * Returns the IP address string "%d.%d.%d.%d"
     * @return raw IP address in a string format
     * @since   JDK1.1
     */
    public String getHostAddress() {	
         return ((address >>> 24) & 0xFF) + "." +
                ((address >>> 16) & 0xFF) + "." +
                ((address >>>  8) & 0xFF) + "." +
                ((address >>>  0) & 0xFF);
     }
 

    /**
     * Returns a hashcode for this IP address.
     *
     * @return  a hash code value for this IP address. 
     * @since   JDK1.0
     */
    public int hashCode() {
	return address;
    }

    /**
     * Compares this object against the specified object.
     * The result is <code>true</code> if and only if the argument is 
     * not <code>null</code> and it represents the same IP address as 
     * this object. 
     * <p>
     * Two instances of <code>InetAddress</code> represent the same IP 
     * address if the length of the byte arrays returned by 
     * <code>getAddress</code> is the same for both, and each of the 
     * array components is the same for the byte arrays. 
     *
     * @param   obj   the object to compare against.
     * @return  <code>true</code> if the objects are the same;
     *          <code>false</code> otherwise.
     * @see     java.net.InetAddress#getAddress()
     * @since   JDK1.0
     */
    public boolean equals(Object obj) {
	return (obj != null) && (obj instanceof InetAddress) &&
	    (((InetAddress)obj).address == address);
    }

    /**
     * Converts this IP address to a <code>String</code>.
     *
     * @return  a string representation of this IP address.
     * @since   JDK1.0
     */
    public String toString() {
	return getHostName() + "/" + getHostAddress();
    }

    /* 
     * Cached addresses - our own litle nis, not! 
     *
     * Do not purge cache of numerical IP addresses, since 
     * duplicate dynamic DNS name lookups can leave the system
     * vulnerable to hostname spoofing attacks.  Once a hostname
     * has been looked up in DNS and entered into the Java cache, 
     * from then on, the hostname is translated to IP address only
     * via the cache.
     */
    static Hashtable	    addressCache = new Hashtable();
    static InetAddress	    unknownAddress;
    static InetAddress	    anyLocalAddress;
    static InetAddress      localHost;
    static InetAddress[]    unknown_array; // put THIS in cache
    
    /* 
     * generic localHost to give back to applets 
     * - private so not API delta
     */
    private static InetAddress      loopbackHost;
    
    static {
	unknownAddress = new InetAddress();
	anyLocalAddress = new InetAddress();
	makeAnyLocalAddress(anyLocalAddress);
	byte[] IP = new byte[4];
	IP[0] = 0x7F;
	IP[1] = 0x00;
	IP[2] = 0x00;
	IP[3] = 0x01;
	loopbackHost = new InetAddress("localhost", IP);

	/* find the local host name */
	try {
	    localHost = new InetAddress();
	    localHost.hostName = getLocalHostName();
	    /* we explicitly leave the address of the local host
	     * uninitialized.  A DNS lookup in this, the static
	     * initializer, will cause a machine disconnected
	     * from the network to hang - it'll be trying to query
	     * a DNS server that isn't there.
	     *
	     * Instead, we just get the hostname of the local host.
	     * The native code for this just calls gethostname()
	     * which should be pretty innocuous - it shouldn't try
	     * to contact a DNS server.  If any application
	     * calls InetAddress.getLocalHost(), we initialize
	     * the local host's address there if not already initialized.
	     *
	     * Note that for this to work it is also essential that
	     * the localHost InetAddress is _NOT_ put into the address cache
	     * here in the static initializer (which happens if we call
	     * getByName() from the static initializer).  It _IS_ OK
	     * to put it in the addressCache after initialization.
	     *
	     * The unitialized state of the localHost's address is -1,
	     * or IP address 255.255.255.255 which we know cannot be
	     * a legal host address.
	     */
	    localHost.address = -1;
	} catch (Exception ex) { /* this shouldn't happen */
	    localHost = unknownAddress;
    }

	/* cache the name/address pair "0.0.0.0"/0.0.0.0 */
	String unknownByAddr = new String("0.0.0.0");
	unknown_array = new InetAddress[1];
	unknown_array[0] = new InetAddress(unknownByAddr, unknownAddress.getAddress());
	addressCache.put(unknownByAddr, unknown_array);
    }

    /**
     * Determines the IP address of a host, given the host's name. The 
     * host name can either be a machine name, such as 
     * "<code>java.sun.com</code>", or a string representing its IP 
     * address, such as "<code>206.26.48.100</code>". 
     *
     * @param      host   the specified host, or <code>null</code> for the
     *                    local host.
     * @return     an IP address for the given host name.
     * @exception  UnknownHostException  if no IP address for the
     *               <code>host</code> could be found.
     * @since      JDK1.0
     */
    public static InetAddress getByName(String host)
	throws UnknownHostException {
	if (host == null || host.length() == 0) {
	    return loopbackHost;
	}

        if (!Character.isDigit(host.charAt(0))) {
	    return getAllByName0(host)[0];
	} else {
	    /* The string (probably) represents a numerical IP address.
	     * If it is not really a numeric address, we treat it as a hostname,
	     * and lookup that way.
	     */
	    InetAddress addr = parseNumericAddress(host);
	    if (addr == null)
	        return getAllByName0(host)[0];
	    return addr;
    }
    }


    private static InetAddress parseNumericAddress (String host) throws UnknownHostException
    {
	    /*
	     * Parse a numeric address into an int.
	     * Leave hostName null, don't cache.  If it isn't an IP address,
	     * (i.e., not "%d.%d.%d.%d") or if any element > 0xFF,
	     * returns null so that caller can look up the address as a hostname.
	     * This seems to be 100% compliant to the RFC1123 spec: 
	     * a partial hostname like 3com.domain4 is technically valid.
	     */

	    int IP = 0x00;
	    int hitDots = 0;
	    char[] data = host.toCharArray();

	    for(int i = 0; i < data.length; i++) {
		char c = data[i];
		if (c < 48 || c > 57) { // !digit
		    return null;
		}
		int b = 0x00;
		while(c != '.') {
		    if (c < 48 || c > 57) { // !digit
			return null;
		    }
		    b = b*10 + c - '0';

		    if (++i >= data.length)
			break;
		    c = data[i];
		}
		if(b > 0xFF) { /* bogus - bigger than a byte */
		    return null;
		}
		IP = (IP << 8) + b;
		hitDots++;
	    }

	    if(hitDots != 4 || host.endsWith(".")) {
		return null;
	    }

        InetAddress in = new InetAddress();
        in.address = IP;
        in.hostName = null;

	    return in;
    }


    /** 
     * Determines all the IP addresses of a host, given the host's name. 
     * The host name can either be a machine name, such as 
     * "<code>java.sun.com</code>", or a string representing 
     * its IP address, such as "<code>206.26.48.100</code>". 
     *
     * @param      host   the name of the host.
     * @return     an array of all the IP addresses for a given host name.
     * @exception  UnknownHostException  if no IP address for the
     *               <code>host</code> could be found.
     * @since      JDK1.0
     */
    public static InetAddress getAllByName(String host)[]
	throws UnknownHostException {

	if (host == null || host.length() == 0) {
	    throw new UnknownHostException("empty string");
	}

	if(Character.isDigit(host.charAt(0))) {
	    InetAddress addr = parseNumericAddress(host);
	    if (addr == null)
	        return getAllByName0(host);
	    InetAddress[] ret = new InetAddress[1];
	    ret[0] = addr;
	    return ret;
	} else {
	    return getAllByName0(host);
	}
    }

    private static InetAddress[] getAllByName0 (String host) 
	throws UnknownHostException  {

	/* If it gets here it is presumed to be a hostname */
	/* Cache.get can return: null, unknownAddress, or InetAddress[] */
        Object obj = null;
	Object objcopy = null;

	/* make sure the connection to the host is allowed, before we
	 * give out a hostname
	 */
    SecurityManager security = System.getSecurityManager();
    if (security != null && !security.getInCheck()) {
        security.checkConnect(host, -1);
    }

	synchronized (addressCache) {
	    obj = addressCache.get(host);

	/* If no entry in cache, then do the host lookup */
	
	    if (obj == null) {

        if (debug)
        {
            System.out.println("looking up "+host);
            Thread.dumpStack();
        }

        // Check if the caller is offline...if so, fail without updating the cache.
        if (BrowserAppletFrame.checkOfflineContext())
        {
            if (debug) System.out.println("offline, can't lookup "+host);
            throw new UnknownHostException(host);
        }


		try {
		    /*
		     * Do not put the call to lookup() inside the
		     * constructor.  if you do you will still be
		     * allocating space when the lookup fails.
		     */
		    Object[] lookup = lookupHostByName(host);
		    obj = updateAddressCache(lookup);
		} catch (UnknownHostException e) {
		    obj  = unknown_array;
		}
		addressCache.put(host, obj);
	    }
	} /* end synchronized block */

	if (obj == unknown_array) {
	    /*
	     * We currently cache the fact that a host is unknown.
	     */
	    throw new UnknownHostException(host);
	}

	/* Make a copy of the InetAddress array */
        objcopy = ((InetAddress [])obj).clone();
        
	return (InetAddress [])objcopy;
    }

    /**
     * Adds a new entry to the address cache.  Synchronized on addressCache
     * by callers.
     */
    private static InetAddress[] updateAddressCache (Object[] lookup)
    {
        String host = (String)lookup[0];
        byte[][] byte_array = (byte[][])lookup[1];

	    InetAddress[] addr_array = new InetAddress[byte_array.length];

	    for (int i = 0; i < byte_array.length; i++) {
		byte addr[] = byte_array[i];
		addr_array[i] = new InetAddress(host, addr);

		// It's necessary to hash the addresses because reverse lookups
		// are not always possible.
		addressCache.put(addr_array[i].getHostAddress(), addr_array);
	    }

        addressCache.put(host, addr_array);

	    return addr_array;
    }

    /**
     * Returns the local host.
     *
     * @return     the IP address of the local host.
     * @exception  UnknownHostException  if no IP address for the
     *               <code>host</code> could be found.
     * @since      JDK1.0
     */
    public static InetAddress getLocalHost() throws UnknownHostException {
        if (localHost.equals(unknownAddress)) {
	    throw new UnknownHostException();
        }

	/* If the localhost's address is not initialized yet, initialize
	 * it.  It is no longer initialized in the static initializer 
	 * (see comment there).
	 */

	if (localHost.address == -1) {
        PolicyEngine.assertPermission(PermissionID.NETIO);
	    /* This puts it in the address cache as well */
	    localHost = getAllByName(localHost.hostName)[0];
        PolicyEngine.revertPermission(PermissionID.NETIO);
	}


        /* make sure the connection to the host is allowed: if yes,
	 * return the "real" localHost; if not, return loopback "127.0.0.1"   
	 */
	SecurityManager security = System.getSecurityManager();
	try {
	    if (security != null && !security.getInCheck()) 
		security.checkConnect(localHost.getHostName(), -1);
	} catch (java.lang.SecurityException e) { 
	    return loopbackHost;
	}
	return localHost;
    }
    
    private static native String getLocalHostName() throws UnknownHostException;
    private static native void makeAnyLocalAddress(InetAddress addr);
    private static native Object[] lookupHostByName(String hostname) throws UnknownHostException;
    private static native Object[] lookupHostByAddr(int addr) throws UnknownHostException;
    private static native int getInetFamily();

    static final boolean debug = false;
}

