/*
 * @(#)LocaleData.java	1.12 97/02/24
 *
 * (C) Copyright Taligent, Inc. 1996 - All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - All Rights Reserved
 *
 * Portions copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 *   The original version of this source code and documentation is copyrighted
 * and owned by Taligent, Inc., a wholly-owned subsidiary of IBM. These
 * materials are provided under terms of a License Agreement between Taligent
 * and Sun. This technology is protected by multiple US and International
 * patents. This notice and attribution to Taligent may not be removed.
 *   Taligent is a registered trademark of Taligent, Inc.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * Copyright (C) 1997 Microsoft Corporation. Portions copyright (C) 1993-1997 Sun Microsystems, Inc & Taligent, Inc & IBM Corp. All Rights Reserved.
 */

package java.text.resources;

import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Hashtable;
import java.util.Enumeration;

import java.io.InputStream;
import java.io.FileNotFoundException;
import com.ms.ui.resource.Win32ResourceDecoder;
import com.ms.ui.resource.ResourceFormattingException;

/**
 * An internal data warehouse for information (locale elements) needed to
 * implement locale dependent formatting for the standard set of locales
 * The current implementation utilizes a class LocaleElements which is
 * created from plain text source file(s) by a small compiler (LocaleTool)
 *
 * Example of use:
 * <pre>
 *  Locale desiredLocale = new Locale("DA","DK","");
 *  ResourceBundle resource = ResourceBundle.load("java.util.LocaleElements",
 *                                    desiredLocale);
 *  // Pass one of the following:
 *  "LocaleString",      // locale id based on iso codes
 *  "LocaleID",          // Windows id
 *  "ShortLanguage",     // iso-3 abbrev lang name
 *  "ShortCountry",      // iso-3 abbrev country name
 *  "Languages",         // language names
 *  "Countries",         // country names
 *  "MonthNames",        // ARRAY january
 *  "MonthAbbreviations",   // ARRAY abb january
 *  "DayNames",          // ARRAY Monday
 *  "DayAbbreviations",  // ARRAY abb Monday
 *  "AmPmMarkers",       // ARRAY am marker
 *  "Eras",              // era strings
 *  "NumberPatterns",    // ARRAY decimal pattern
 *  "CurrencyElements",  // ARRAY local currency symbol
 *  "DateTimePatterns",  // ARRAY full time pattern
 *  "DateTimeElements",  // ARRAY first day of week
 *  "CollationElements", // collation order
 *
 *  String myString string = resource.getString(LocaleString);
 *  // or
 *  String[] myStrings strings = resource.getStringArray(LocaleString);
 * </pre>
 * @author Asmus Freytag
 * @author Mark Davis
 * @version 1.12 02/24/97
 */

public class LocaleData extends ResourceBundle 
{

    // each subclass MUST call Init() in its constructor.
    public Object handleGetObject(String key)
    {
        return localeKeys.get(key);
    }

    // use to get the list of locales that have this data
    public static Locale[] getAvailableLocales(String key)
    {
        return localeList;  // hard-coded for now
    }

    /**
     * Implementation of ResourceBundle.getKeys.
     */
    public Enumeration getKeys() {
        return localeKeys.keys();
    }

    // given string data that matches the keys, sets up the hash
    // table for quick access.
    protected void init(String[] data) {
        if (data.length != keys.length) {
            System.out.println("Data length (" + data.length +
                               ") != key length (" + keys.length + ")");
            throw new ArrayIndexOutOfBoundsException();
        }
        // set ups a hashtable for the locale keys
        localeKeys = new Hashtable();

        int oldIndex = 0;
        int i;
        for(i = 1; i < data.length; i++ )
        {
            if (!keys[i].equals(keys[oldIndex])) {
                storeStrings(data,oldIndex, i-1);
                oldIndex = i;
            }
        }
        storeStrings(data,oldIndex, i-1);
    }

    // ========== privates ==========

    private Hashtable localeKeys;

    /*
     * Splits the string into pieces that it returns in an array.
     * @param source string to split
     * @separator char to use to split. NOTE: two separators in a row,
     * or one at the end or start, will produce a null string; e.g.
     * ";a;b;" -> {"", "a", "b", ""}
     */
    private static String[] split(String source, char separator) {
        String[] temp = new String[200]; // TODO remove limited length
        int oldOffset = 0;
        int i;
        for (i = 0; oldOffset < source.length(); ++i) {
            int newOffset = source.indexOf(separator, oldOffset);
            if (newOffset == -1)
                newOffset = source.length();
            temp[i] = source.substring(oldOffset, newOffset);
            oldOffset = newOffset + 1;
        }
        String[] result = new String[i];
        System.arraycopy(temp,0,result,0,i);
        return result;
    }
    // stores either a single string, or an array (in case of multiple keys)
    private void storeStrings(String[] data, int start, int end) {
        if (keys[start].equals("Eras")
            || keys[start].equals("Languages")
            || keys[start].equals("Countries")) {
            String[] strings = split(data[start], ';');
            // workaround for bug
            for (int i = 0; i < strings.length; ++i)
                strings[i] = strings[i].trim();
            // break up "_'
            if (keys[start].equals("Eras"))
                localeKeys.put(keys[start], strings);
            else {
                String[][] pairedStrings = new String[strings.length][2];
                for (int i = 0; i < strings.length; ++i)
                    pairedStrings[i] = split(strings[i],'_');
                localeKeys.put(keys[start], pairedStrings);
            }
        } else if (start == end) { // store single string
            localeKeys.put(keys[start], data[start]);
        } else {
            String[] temp = new String[end - start + 1];
            System.arraycopy(data, start, temp, 0, end - start + 1);
            localeKeys.put(keys[start], temp);
        }
    }

    private static String keys[];
    private static Locale localeList[];

	static
	{
		Class cl=LocaleData.class;
		InputStream is=cl.getResourceAsStream("LocaleData.res");
		if(is==null)
			System.out.println("Can not load resources for LocaleData");
		else
		{
			keys = new String[75];
			localeList = new Locale[50];
			Win32ResourceDecoder resources = null;
			int str = 1;
			try
			{
				resources=new Win32ResourceDecoder( is );
				for(int i=0; i<75; i++)
					keys[i] = resources.getString(str++);
			}
			catch(FileNotFoundException fnfe)
			{
				System.out.println("file not found!!");
				resources = null;
			}
			catch(ResourceFormattingException rfe)
			{
				System.out.println("Resource formatting exception!!");
				resources = null;
			}

			if(resources != null)
			{
				for(int i=0; i<50; i++)
				{
					String a = "";
					String b = "";
					String c = "";
					try
					{
						a = resources.getString(str++);
					}
					catch(ResourceFormattingException rfe) {}

					try
					{
						b = resources.getString(str++);
					}
					catch(ResourceFormattingException rfe) {}

					try
					{
						c = resources.getString(str++);
					}
					catch(ResourceFormattingException rfe) {}

					localeList[i] = new Locale(a,b,c);
				}
			}
		}
    }
}
